import boto3
import time
import os
from datetime import datetime
from tibanna import create_logger
from tibanna.vars import AMI_PER_REGION


logger = create_logger(__name__)


class AMI(object):

    BASE_AMI = 'ami-0885b1f6bd170450c'  # ubuntu 20.04 for us-east-1
    BASE_REGION = 'us-east-1'
    USERDATA_DIR = os.path.dirname(os.path.abspath(__file__))
    USERDATA_FILE = os.path.join(USERDATA_DIR, 'create_ami_userdata')
    AMI_NAME = 'tibanna-ami-' + datetime.strftime(datetime.today(), '%Y%m%d')  # e.g tibanna-ami-20201113

    def __init__(self, base_ami=None, base_region=None, userdata_file=None, ami_name=None):
        if base_ami:
            self.BASE_AMI = base_ami
        if base_region:
            self.BASE_REGION = base_region
        if userdata_file is not None:
            self.USERDATA_FILE = userdata_file
        if ami_name:
            self.AMI_NAME = ami_name

    @staticmethod
    def launch_instance_for_tibanna_ami(keyname, userdata_file, base_ami):

        launch_args = {'ImageId': base_ami,
                       'InstanceType': 't3.micro',
                       'MaxCount': 1,
                       'MinCount': 1,
                       'TagSpecifications': [{'ResourceType': 'instance',
                                              'Tags': [{"Key": "Name", "Value": "tibanna_ami"}]}]}
        if userdata_file:
            with open(userdata_file, 'r') as f:
                userdata_str = f.read()
            launch_args.update({'UserData': userdata_str})

        if keyname:
            launch_args.update({'KeyName': keyname})

        logger.debug("launch_args=" + str(launch_args))
        ec2 = boto3.client('ec2')
        res = ec2.run_instances(**launch_args)
        logger.debug("response from EC2 run_instances :" + str(res) + '\n\n')
        instance_id = res['Instances'][0]['InstanceId']

        return instance_id

    def create_ami_for_tibanna(self, keyname=None, make_public=False, replicate=False):
        return self.create_ami(keyname=keyname, userdata_file=self.USERDATA_FILE,
                               base_ami=self.BASE_AMI, ami_name=self.AMI_NAME,
                               make_public=make_public, base_region=self.BASE_REGION, replicate=replicate)

    @staticmethod
    def replicate_ami(*, ami_name, ami_id, source_region='us-east-1',
                      target_regions=None,
                      make_public=True):
        """ Replicates the given ami_id from the source region into the target region(s).

            Params:
                ami_name (str): Name to use with AMI, typically something like tibanna-ami-20211025
                ami_id (str): The AMI ID from AWS for the source image
                source_region (str): The region the source AMI was created in
                target_regions (str[]): List of regions to replicate the source AMI into
                                        Leave as None to use all regions in AMI_PER_REGION
                make_public (bool, default True): whether or not this AMI should be public

            Raises ClientError if a boto3 call fails.
            Returns an AMI_PER_REGION mapping
        """
        if not target_regions:
            target_regions = [r for r in AMI_PER_REGION.keys() if r != source_region]

        # Create sessions in each target region and copy the AMI into it
        # If this AMI is to be publicly available, sleep for 5 mins to allow
        # the AMI to be registered, then modify attribution to public.
        ami_per_region = {}
        for region in target_regions:
            region_session = boto3.client('ec2', region_name=region)
            response = region_session.copy_image(
                Name=ami_name,
                Description=f'{ami_name} replicated from {source_region}',
                SourceImageId=ami_id,
                SourceRegion=source_region
            )
            new_image_id = response['ImageId']
            logger.info(f'Copied {ami_name} from {source_region} to {region}'
                        f' under new AMI ID {new_image_id}')
            ami_per_region[region] = new_image_id

        if make_public:
            logger.info('Provisioning PUBLIC AMIs - sleeping 5 mins, ctrl-c now if unintended')
            time.sleep(5 * 60)
            for region, image_id in ami_per_region.items():
                region_session = boto3.client('ec2', region_name=region)  # re-create since its been some time
                region_session.modify_image_attribute(ImageId=image_id,
                                                      LaunchPermission={'Add': [{'Group': 'all'}]})
        else:
            logger.info(f'Provisioning private AMIs')

        logger.info(f'Provisioned {ami_per_region}')
        return ami_per_region

    @classmethod
    def create_ami(cls, keyname=None, userdata_file=USERDATA_FILE,
                   base_ami=BASE_AMI,
                   ami_name=AMI_NAME,
                   make_public=False,
                   replicate=False,
                   base_region='us-east-1'):
        """ Helper function that creates the Tibanna AMI from a base image. """
        if not userdata_file:
            logger.info("no userdata.. no need to launch an instance.. just copying image")
            ec2 = boto3.client('ec2')
            try:
                res_copy = ec2.copy_image(Name=ami_name, SourceImageId=base_ami, SourceRegion=base_region)
            except:
                raise Exception("Failed to copy image")

            # I tried 5 min - it's not enough and it fails at the next step.
            logger.info("waiting for 10min for the image to be created..")
            time.sleep(10 * 60)

            new_image_id = res_copy['ImageId']

            if make_public:
                ec2.modify_image_attribute(ImageId=new_image_id,
                                           LaunchPermission={'Add': [{'Group': 'all'}]})
            return new_image_id

        # Launch an instance with base AMI
        try:
            instance_id = AMI.launch_instance_for_tibanna_ami(keyname, userdata_file, base_ami)
            logger.debug("instance_id=" + instance_id)
        except:
            raise Exception("Failed to launch an instance")

        logger.info("waiting for 10min for the instance to install everything and reboot..")
        time.sleep(10 * 60)

        # Create an image from the instance
        try:
            create_image_args = {'InstanceId': instance_id, 'Name':  ami_name}
            ec2 = boto3.client('ec2')
            logger.info("creating an image...")
            res_create = ec2.create_image(**create_image_args)
        except:
            raise Exception("Failed to create an image")

        logger.info("waiting for 10min for the image to be created..")
        time.sleep(10 * 60)

        # Terminate instance once enough time has passed that we are certain AMI creation
        # should have completed
        try:
            ec2.terminate_instances(InstanceIds=[instance_id])
        except:
            raise Exception("Failed to terminate the instance")

        new_image_id = res_create['ImageId']

        # Make new base image public
        if make_public:
            ec2.modify_image_attribute(ImageId=new_image_id,
                                       LaunchPermission={'Add': [{'Group': 'all'}]})

        # Replicate the image across regions as desired
        if replicate:
            cls.replicate_ami(ami_name=ami_name, ami_id=new_image_id, source_region=base_region,
                              make_public=make_public)

        return new_image_id



import copy


class SerializableObject(object):
    def __init__(self):
        pass

    def update(self, **kwargs):
        for k, v in kwargs.items():
            setattr(self, k, v)

    def as_dict(self):
        # use deepcopy so that changing this dictionary later won't affect the SerializableObject
        d = copy.deepcopy(self.__dict__)
        for k in list(d.keys()):
            if d[k] is None:
                del d[k]
        # recursive serialization
        for k, v in d.items():
            if isinstance(v, SerializableObject):
                d[k] = v.as_dict()
            if isinstance(v, list):
                for i, e in enumerate(v):
                    if isinstance(e, SerializableObject):
                        d[k][i] = e.as_dict()
            if isinstance(v, dict):
                for l, w in v.items():
                    if isinstance(w, SerializableObject):
                        d[k][l] = w.as_dict()
        return d



import random
import string
import boto3
import os
import mimetypes
from uuid import uuid4, UUID
from . import create_logger
from .vars import (
    _tibanna,
    EXECUTION_ARN,
    AWS_REGION
)


logger = create_logger(__name__)


def _tibanna_settings(settings_patch=None, force_inplace=False, env=''):
    tbn = {"run_id": str(uuid4()),
           "env": env,
           "url": '',
           'run_type': 'generic',
           'run_name': '',
           }
    in_place = None
    if force_inplace:
        if not settings_patch.get(_tibanna):
            settings_patch[_tibanna] = {}
    if settings_patch:
        in_place = settings_patch.get(_tibanna, None)
        if in_place is not None:
            tbn.update(in_place)
        else:
            tbn.update(settings_patch)

    # generate run name
    if not tbn.get('run_name'):
        # aws doesn't like / in names
        tbn['run_name'] = "%s_%s" % (tbn['run_type'].replace('/', '-'), tbn['run_id'])

    if in_place is not None:
        settings_patch[_tibanna] = tbn
        return settings_patch
    else:
        return {_tibanna: tbn}


def randomize_run_name(run_name, sfn):
    arn = EXECUTION_ARN(run_name, sfn)
    client = boto3.client('stepfunctions', region_name=AWS_REGION)
    try:
        response = client.describe_execution(
                executionArn=arn
        )
        if response:
            if len(run_name) > 36:
                try:
                    UUID(run_name[-36:])
                    run_name = run_name[:-37]  # remove previous uuid
                except:
                    pass
            run_name += '-' + str(uuid4())

    except Exception:
        pass
    return run_name


# random string generator
def randomword(length):
    choices = string.ascii_lowercase + string.ascii_uppercase + string.digits
    return ''.join(random.choice(choices) for i in range(length))


def create_jobid():
    return randomword(12)    # date+random_string


def create_tibanna_suffix(suffix, usergroup):
    if usergroup and suffix:
        function_name_suffix = usergroup + '_' + suffix
    elif suffix:
        function_name_suffix = suffix
    elif usergroup:
        function_name_suffix = usergroup
    else:
        function_name_suffix = ''

    if function_name_suffix:
        return '_' + function_name_suffix
    else:
        return ''


def read_s3(bucket, object_name):
    response = boto3.client('s3').get_object(Bucket=bucket, Key=object_name)
    logger.debug("response_from_read_s3:" +  str(response))
    return response['Body'].read().decode('utf-8', 'backslashreplace')


def does_key_exist(bucket, object_name, quiet=False):
    try:
        file_metadata = boto3.client('s3').head_object(Bucket=bucket, Key=object_name)
    except Exception as e:
        if not quiet:
            print("object %s not found on bucket %s" % (str(object_name), str(bucket)))
            print(str(e))
        return False
    return file_metadata


def upload(filepath, bucket, prefix='', public=True, encrypt_s3_upload=False, kms_key_id=None):
    """ Uploads a file to S3 under a prefix.
        The original directory structure is removed
        and only the filename is preserved.
        If filepath is none, upload an empty file with prefix
        itself as key
        If encrypt_s3_upload is True, default Server-Side encryption is enabled
        If kms_key_id is True as well, Server-Side encryption with the given key_id is enabled
    """
    if public:
        acl = 'public-read'
    else:
        acl = 'private'
    s3 = boto3.client('s3')
    if encrypt_s3_upload:
        upload_extra_args = {'ServerSideEncryption': 'aws:kms'}
        if kms_key_id:
            upload_extra_args['SSEKMSKeyId'] = kms_key_id
    else:
        upload_extra_args = {}
    if filepath:
        dirname, filename = os.path.split(filepath)
        key = os.path.join(prefix, filename)
        logger.debug('filepath=%s, filename=%s, key=%s' % (filepath, filename, key))
        content_type = mimetypes.guess_type(filename)[0]
        if content_type is None:
            content_type = 'binary/octet-stream'
        upload_extra_args.update({'ContentType': content_type})
        try:
            upload_extra_args.update({'ACL': acl})
            s3.upload_file(filepath, bucket, key, ExtraArgs=upload_extra_args)
        except Exception as e:
            upload_extra_args.update({'ACL': 'private'})
            s3.upload_file(filepath, bucket, key, ExtraArgs=upload_extra_args)
    else:
        try:
            s3.put_object(Body=b'', Bucket=bucket, Key=prefix, ACL=acl, **upload_extra_args)
        except Exception as e:
            s3.put_object(Body=b'', Bucket=bucket, Key=prefix, ACL='private', **upload_extra_args)


def put_object_s3(content, key, bucket, public=True, encrypt_s3_upload=False, kms_key_id=None):
    if public:
        acl = 'public-read'
    else:
        acl = 'private'
    s3 = boto3.client('s3')
    content_type = mimetypes.guess_type(key)[0]
    if content_type is None:
        content_type = 'binary/octet-stream'
    if encrypt_s3_upload:
        upload_extra_args = {'ServerSideEncryption': 'aws:kms'}
        if kms_key_id:
            upload_extra_args['SSEKMSKeyId'] = kms_key_id
    else:
        upload_extra_args = {}
    try:
        s3.put_object(Body=content.encode('utf-8'), Bucket=bucket, Key=key, ACL=acl,
                      ContentType=content_type, **upload_extra_args)
    except Exception as e:
        s3.put_object(Body=content.encode('utf-8'), Bucket=bucket, Key=key, ACL='private',
                      ContentType=content_type, **upload_extra_args)


def retrieve_all_keys(prefix, bucket):
    s3 = boto3.client('s3')
    ContinuationToken=''
    keylist = []
    list_input = {'Bucket': bucket, 'Prefix': prefix}
    while(True):
        if ContinuationToken:
            list_input.update({'ContinuationToken': ContinuationToken})
        res = s3.list_objects_v2(**list_input)
        keylist.extend([str(rc['Key']) for rc in res.get('Contents', [])])
        if 'IsTruncated' in res and res['IsTruncated']:
            ContinuationToken=res['NextContinuationToken']
        else:
            break
    return keylist


def delete_keys(keylist, bucket):
    s3 = boto3.client('s3')
    max_n = 1000  # limit for number of objects to be deleted together
    i_curr = 0
    while(i_curr< len(keylist)):
        i_next = min(i_curr + max_n, len(keylist))
        object_list = [{'Key': k} for k in keylist[i_curr:i_next]]
        s3.delete_objects(
            Bucket=bucket,
            Delete={
                'Objects': object_list, 'Quiet': True
            }
        )
        i_curr = i_next



"""Version information."""
import tomlkit
import os
from pathlib import Path


def _get_project_meta():
    try:
        toml_path = Path(__file__).parent.parent.joinpath('pyproject.toml')
        with open(toml_path) as pyproject:
            file_contents = pyproject.read()

        return tomlkit.parse(file_contents)['tool']['poetry']
    except:
        return "version_not_found"
    

pkg_meta = _get_project_meta()

# Lambdas are deployed with the TIBANNA_VERSION env variable and get version information from there, as
# they can't parse the pyproject.toml. For all other use cases, fall back to the pyproject.toml version
__version__ = os.environ.get('TIBANNA_VERSION', False) or str(pkg_meta['version'])




import boto3
import json
import random
from . import create_logger
from .vars import (
    DYNAMODB_TABLE,
    AWS_ACCOUNT_NUMBER,
    AWS_REGION,
    LAMBDA_TYPE,
    SFN_TYPE,
    RUN_TASK_LAMBDA_NAME,
    CHECK_TASK_LAMBDA_NAME,
    UPDATE_COST_LAMBDA_NAME,
    S3_ENCRYT_KEY_ID,
)


logger = create_logger(__name__)


class IAM(object):

    account_id = AWS_ACCOUNT_NUMBER
    region = AWS_REGION
    lambda_type = LAMBDA_TYPE  # lambda_type : '' for unicorn, 'pony' for pony, 'zebra' for zebra
    sfn_type = SFN_TYPE  # sfn type : 'unicorn' for unicorn, 'pony' for pony, 'zebra' for zebra
    run_task_lambda_name = RUN_TASK_LAMBDA_NAME
    check_task_lambda_name = CHECK_TASK_LAMBDA_NAME
    update_cost_lambda_name = UPDATE_COST_LAMBDA_NAME

    def __init__(self, user_group_tag, bucket_names='', no_randomize=True):
        """policy prefix for user group
        lambda_type : '' for unicorn, 'pony' for pony, 'zebra' for zebra
        example>
          user_group_tag : default
          user_group_name : default_3465
          tibanna_policy_prefix : tibanna_default_3465 / tibanna_pony_default_3465
          prefix : tibanna_ / tibanna_pony_
        """
        if self.lambda_type:
            self.prefix = 'tibanna_' + self.lambda_type + '_'
        else:
            self.prefix = 'tibanna_'

        # user group name & tibanna_policy_prefix
        self.user_group_tag = user_group_tag
        # add rangom tag to avoid attempting to overwrite a previously created and deleted policy and silently failing.
        if no_randomize:
            self.user_group_name = self.user_group_tag
        else:
            random_tag = str(int(random.random() * 10000))
            self.user_group_name = self.user_group_tag + '_' + random_tag
        self.tibanna_policy_prefix = self.prefix + self.user_group_name
        self.tibanna_sfn_name = self.prefix + self.sfn_type + '_' + self.user_group_name

        # bucket names
        self.bucket_names = bucket_names

        # iam client/resource
        self.client = boto3.client('iam')
        self.iam = boto3.resource('iam')

    @property
    def lambda_names(self):
        return [self.run_task_lambda_name, self.check_task_lambda_name, self.update_cost_lambda_name]

    @property
    def iam_group_name(self):
        return self.tibanna_policy_prefix

    @property
    def policy_types(self):
        return ['bucket', 'termination', 'list', 'cloudwatch', 'passrole', 'lambdainvoke',
                'cloudwatch_metric', 'cw_dashboard', 'dynamodb', 'ec2_desc',
                'executions', 'pricing', 'vpc', 'kms']

    def policy_arn(self, policy_type):
        return 'arn:aws:iam::' + self.account_id + ':policy/' + self.policy_name(policy_type)

    def kms_key_arn(self):
        _id = S3_ENCRYT_KEY_ID
        if not _id:
            _id = 'invalid'  # build a dummy arn in the case that we are not using KMS
        return 'arn:aws:kms:' + self.region + ':' + self.account_id + ':key/' + _id

    def policy_suffix(self, policy_type):
        suffices = {'bucket': 'bucket_access',
                    'termination': 'ec2_termination',
                    'list': 'list_instanceprofiles',
                    'cloudwatch': 'cloudwatchlogs',
                    'passrole': 'iam_passrole_s3',
                    'lambdainvoke': 'lambdainvoke',
                    'cloudwatch_metric': 'cw_metric',
                    'cw_dashboard': 'cw_dashboard',
                    'dynamodb': 'dynamodb',
                    'ec2_desc': 'ec2_desc',
                    'pricing': 'pricing',
                    'executions': 'executions',
                    'vpc': 'vpc_access',
                    'kms': 'kms_key_for_s3'}
        if policy_type not in suffices:
            raise Exception("policy %s must be one of %s." % (policy_type, str(self.policy_types)))
        return suffices[policy_type]

    def policy_name(self, policy_type):
        return self.tibanna_policy_prefix + '_' + self.policy_suffix(policy_type)

    def policy_definition(self, policy_type):
        definitions = {'bucket': self.policy_bucket_access,
                       'termination': self.policy_terminate_instances,
                       'list': self.policy_list_instanceprofiles,
                       'cloudwatch': self.policy_cloudwatchlogs,
                       'passrole': self.policy_iam_passrole_s3,
                       'lambdainvoke': self.policy_lambdainvoke,
                       'cloudwatch_metric': self.policy_cloudwatch_metric,
                       'cw_dashboard': self.policy_cw_dashboard,
                       'dynamodb': self.policy_dynamodb,
                       'ec2_desc': self.policy_ec2_desc_policy,
                       'pricing': self.policy_pricing,
                       'executions': self.policy_executions,
                       'vpc': self.policy_vpc_access,
                       'kms': self.policy_kms_access}
        if policy_type not in definitions:
            raise Exception("policy %s must be one of %s." % (policy_type, str(self.policy_types)))
        return definitions[policy_type]

    @property
    def role_types(self):
        return ['ec2', 'stepfunction'] + self.lambda_names

    def role_suffix(self, role_type):
        suffices = {'ec2': 'for_ec2',
                    'stepfunction': 'states'}
        suffices.update({_: _ for _ in self.lambda_names})
        if role_type not in suffices:
            raise Exception("role_type %s must be one of %s." % (role_type, str(self.role_types)))
        return suffices[role_type]

    def role_name(self, role_type):
        return self.tibanna_policy_prefix + '_' + self.role_suffix(role_type)

    def role_service(self, role_type):
        services = {'ec2': 'ec2',
                    'stepfunction': 'states'}
        services.update({_: 'lambda' for _ in self.lambda_names})
        if role_type not in services:
            raise Exception("role_type %s must be one of %s." % (role_type, str(self.role_types)))
        return services[role_type]

    @property
    def policy_arn_list_for_role(self):
        """ This function specifies the permissions given to each service/lambda function. """
        base = ['bucket']  # policies everyone needs
        if S3_ENCRYT_KEY_ID:  # if we have a value, this key is valid and we must add perms
            base.append('kms')
        # returns a dictionary with role_type as keys
        # adding vpc access to only check_task since run_task has full ec2 access
        run_task_custom_policy_types = base + ['list', 'cloudwatch', 'passrole', 'dynamodb',
                                               'executions', 'cw_dashboard']
        check_task_custom_policy_types = base + ['cloudwatch_metric', 'cloudwatch', 'ec2_desc',
                                                 'termination', 'dynamodb', 'pricing', 'vpc']
        update_cost_custom_policy_types = base + ['executions', 'dynamodb', 'pricing', 'vpc']
        arnlist = {'ec2': [self.policy_arn(_) for _ in base + ['cloudwatch_metric', 'ec2_desc']] +
                          ['arn:aws:iam::aws:policy/AmazonEC2ContainerRegistryReadOnly'] +
                          ['arn:aws:iam::aws:policy/CloudWatchAgentServerPolicy'],
                   # 'stepfunction': [self.policy_arn(_) for _ in ['lambdainvoke']],
                   'stepfunction': ['arn:aws:iam::aws:policy/service-role/AWSLambdaRole'],
                   self.run_task_lambda_name: [self.policy_arn(_) for _ in run_task_custom_policy_types] +
                                              ['arn:aws:iam::aws:policy/AmazonEC2FullAccess'],
                   self.check_task_lambda_name: [self.policy_arn(_) for _ in check_task_custom_policy_types],
                   self.update_cost_lambda_name: [self.policy_arn(_) for _ in update_cost_custom_policy_types]}

        return arnlist

    @property
    def instance_profile_name(self):
        return self.role_name('ec2')

    @property
    def policy_bucket_access(self):
        if self.bucket_names:
            resource_list_buckets = ["arn:aws:s3:::" + bn for bn in self.bucket_names]
            resource_list_objects = ["arn:aws:s3:::" + bn + "/*" for bn in self.bucket_names]
        else:
            resource_list_buckets = ["arn:aws:s3:::" + "my-tibanna-test-bucket",
                                     "arn:aws:s3:::" + "my-tibanna-test-input-bucket"]
            resource_list_objects = ["arn:aws:s3:::" + "my-tibanna-test-bucket/*",
                                     "arn:aws:s3:::" + "my-tibanna-test-input-bucket/*"]
        policy_bucket_access = {
            "Version": "2012-10-17",
            "Statement": [
                {
                    "Effect": "Allow",
                    "Action": [
                        "s3:ListBucket"
                    ],
                    "Resource": resource_list_buckets
                },
                {
                    "Effect": "Allow",
                    "Action": [
                        "s3:PutObject",
                        "s3:GetObject",
                        "s3:DeleteObject",
                        "s3:PutObjectAcl",
                        "s3:PutObjectTagging"
                    ],
                    "Resource": resource_list_objects
                }
            ]
        }
        return policy_bucket_access

    @property
    def policy_terminate_instances(self):
        policy = {
            "Version": "2012-10-17",
            "Statement": [
                {
                    "Effect": "Allow",
                    "Action": "ec2:TerminateInstances",
                    "Resource": "*"
                }
            ]
        }
        return policy

    @property
    def policy_vpc_access(self):
        policy = {
            "Version": "2012-10-17",
            "Statement": [
                {
                    "Effect": "Allow",
                    "Action": [
                        "ec2:DescribeNetworkInterfaces",
                        "ec2:CreateNetworkInterface",
                        "ec2:AttachNetworkInterface",
                        "ec2:DeleteNetworkInterface"
                    ],
                    "Resource": "*"
                }
            ]
        }
        return policy

    @property
    def policy_kms_access(self):
        """ Returns a dictionary representing a policy granting some limited
            permissions needed to upload/download files to encrypted S3 buckets
            given an S3_ENCRYPT_KEY_ID.
        """
        return {
            'Version': '2012-10-17',
            'Statement': {
                'Effect': 'Allow',
                'Action': [
                    'kms:Encrypt',
                    'kms:Decrypt',
                    'kms:ReEncrypt*',
                    'kms:GenerateDataKey*',
                    'kms:DescribeKey'
                ],
                'Resource': [
                    self.kms_key_arn()
                ]
            }
        }

    @property
    def policy_list_instanceprofiles(self):
        policy_list_instanceprofiles = {
            "Version": "2012-10-17",
            "Statement": [
                {
                    "Sid": "Stmt1478801433000",
                    "Effect": "Allow",
                    "Action": [
                        "iam:ListInstanceProfiles"
                    ],
                    "Resource": [
                        "*"
                    ]
                }
            ]
        }
        return policy_list_instanceprofiles

    @property
    def policy_cloudwatchlogs(self):
        policy_cloudwatchlogs = {
            "Version": "2012-10-17",
            "Statement": [
                {
                    "Action": [
                        "logs:CreateLogGroup",
                        "logs:CreateLogStream",
                        "logs:PutLogEvents"
                    ],
                    "Resource": "arn:aws:logs:*:*:*",
                    "Effect": "Allow"
                }
            ]
        }
        return policy_cloudwatchlogs

    @property
    def policy_iam_passrole_s3(self):
        role_resource = ['arn:aws:iam::' + self.account_id + ':role/' + self.tibanna_policy_prefix + '_for_ec2']
        policy_iam_passrole_s3 = {
            "Version": "2012-10-17",
            "Statement": [
                {
                    "Sid": "Stmt1478801396000",
                    "Effect": "Allow",
                    "Action": [
                        "iam:PassRole"
                    ],
                    "Resource": role_resource
                }
            ]
        }
        return policy_iam_passrole_s3

    @property
    def policy_lambdainvoke(self):
        function_arn_prefix = 'arn:aws:lambda:' + self.region + ':' + self.account_id + ':function/'
        resource = [function_arn_prefix + ln + '_' + self.tibanna_policy_prefix for ln in self.lambda_names]
        policy = {
            "Version": "2012-10-17",
            "Statement": [
                {
                    "Effect": "Allow",
                    "Action": [
                        "lambda:InvokeFunction"
                    ],
                    "Resource": resource
                }
            ]
        }
        return policy

    @property
    def policy_executions(self):
        execution_arn_prefix = 'arn:aws:states:' + self.region + ':' + self.account_id + ':stateMachine:'
        sfn_arn_prefix = 'arn:aws:states:' + self.region + ':' + self.account_id + ':execution:'
        resources = [execution_arn_prefix + self.tibanna_sfn_name,
                    execution_arn_prefix + self.tibanna_sfn_name + '_costupdater',
                    sfn_arn_prefix + self.tibanna_sfn_name + ':*',
                    sfn_arn_prefix + self.tibanna_sfn_name + '_costupdater:*']
        policy = {
            "Version": "2012-10-17",
            "Statement": [
                {
                    "Effect": "Allow",
                    "Action": [
                        "states:StartExecution",
                        "states:StopExecution",
                        "states:ListExecutions",
                        "states:DescribeExecution",
                        "states:GetExecutionHistory",
                        "states:DescribeStateMachineForExecution",
                        "states:DescribeStateMachine"
                    ],
                    "Resource": resources
                },
                {
                    "Sid": "VisualEditor1",
                    "Effect": "Allow",
                    "Action": "states:ListStateMachines",
                    "Resource": "*"
                }
            ]
        }
        return policy

    @property
    def policy_cloudwatch_metric(self):
        policy = {
            "Version": "2012-10-17",
            "Statement": [
                {
                    "Effect": "Allow",
                    "Action": [
                        "cloudwatch:PutMetricData",
                        "cloudwatch:GetMetricStatistics"
                    ],
                    "Resource": "*"
                }
            ]
        }
        return policy

    @property
    def policy_cw_dashboard(self):
        policy = {
            "Version": "2012-10-17",
            "Statement": [
                {
                    "Effect": "Allow",
                    "Action": [
                        "cloudwatch:PutDashboard"
                    ],
                    "Resource": "*"
                }
            ]
        }
        return policy

    @property
    def policy_dynamodb(self):
        policy = {
            "Version": "2012-10-17",
            "Statement": [
                {
                    "Effect": "Allow",
                    "Action": [
                        "dynamodb:DescribeTable",
                        "dynamodb:PutItem",
                        "dynamodb:Query",
                        "dynamodb:UpdateItem"
                    ],
                    "Resource": "arn:aws:dynamodb:" + self.region + ":" + self.account_id + ":table/" + DYNAMODB_TABLE
                }
            ]
        }
        return policy

    @property
    def policy_ec2_desc_policy(self):
        policy = {
            "Version": "2012-10-17",
            "Statement": [
                {
                    "Effect": "Allow",
                    "Action": [
                        "ec2:DescribeInstances",
                        "ec2:DescribeInstanceStatus",
                        "ec2:DescribeSpotPriceHistory",
                        "ec2:DescribeSpotInstanceRequests"
                    ],
                    "Resource": "*"
                }
            ]
        }
        return policy

    @property
    def policy_pricing(self):
        policy = {
            "Version": "2012-10-17",
            "Statement": [
                {
                    "Effect": "Allow",
                    "Action": [
                        "pricing:DescribeServices",
                        "pricing:GetAttributeValues",
                        "pricing:GetProducts",
                        "ce:GetCostAndUsage"
                    ],
                    "Resource": "*"
                }
            ]
        }
        return policy

    def role_policy_document(self, service):
        '''service: 'ec2', 'lambda' or 'states' '''
        AssumeRolePolicyDocument = {
            "Version": "2012-10-17",
            "Statement": [
                {
                    "Effect": "Allow",
                    "Principal": {
                        "Service": service + ".amazonaws.com"
                    },
                    "Action": "sts:AssumeRole"
                }
            ]
        }
        return AssumeRolePolicyDocument

    def create_role_robust(self, rolename, roledoc, verbose=False):
        try:
            response = self.client.create_role(
               RoleName=rolename,
               AssumeRolePolicyDocument=roledoc
            )
        except Exception as e:
            if 'EntityAlreadyExists' in str(e):
                try:
                    # first remove
                    self.remove_role(rolename, ignore_errors=False)
                    # recreate
                    response = self.client.create_role(
                       RoleName=rolename,
                       AssumeRolePolicyDocument=roledoc
                    )
                except Exception as e2:
                    raise Exception("Can't create role %s: %s" % (rolename, str(e2)))
        if verbose:
            logger.debug("response from create_role_robust: " + str(response))

    def create_empty_role_for_lambda(self, verbose=False):
        role_policy_doc_lambda = self.role_policy_document('lambda')
        empty_role_name = 'tibanna_lambda_init_role'
        try:
            self.client.get_role(RoleName=empty_role_name)
        except Exception:
            logger.info("creating %s", empty_role_name)
            self.create_role_robust(empty_role_name, json.dumps(role_policy_doc_lambda), verbose)

    def create_role_for_role_type(self, role_type, verbose=False):
        role_policy_doc = self.role_policy_document(self.role_service(role_type))
        self.create_role_robust(self.role_name(role_type), json.dumps(role_policy_doc), verbose)
        role = self.iam.Role(self.role_name(role_type))
        for p_arn in self.policy_arn_list_for_role[role_type]:
            response = role.attach_policy(PolicyArn=p_arn)
            if verbose:
                logger.debug("response from IAM attach_policy :" + str(response))

    def create_user_group(self, verbose=False):
        try:
            response = self.client.create_group(
               GroupName=self.iam_group_name
            )
            if verbose:
                logger.debug("response from IAM create_group :" + str(response))
        except Exception as e:
            if 'EntityAlreadyExists' in str(e):
                # do not actually delete the group, just detach existing policies.
                # deleting a group would require users to be detached from the group.
                self.detach_policies_from_group()
        group = self.iam.Group(self.iam_group_name)
        response = group.attach_policy(
            PolicyArn='arn:aws:iam::aws:policy/service-role/AWSLambdaBasicExecutionRole'
        )
        if verbose:
            logger.debug("response from IAM attach_policy :" + str(response))
        response = group.attach_policy(
            PolicyArn='arn:aws:iam::aws:policy/CloudWatchReadOnlyAccess'
        )
        if verbose:
            logger.debug("response from IAM attach_policy :" + str(response))
        custom_policy_types = ['bucket', 'ec2_desc', 'cloudwatch_metric', 'dynamodb', 'termination', 'pricing', 'executions']
        for pn in [self.policy_name(pt) for pt in custom_policy_types]:
            response = group.attach_policy(
                PolicyArn='arn:aws:iam::' + self.account_id + ':policy/' + pn
            )
            if verbose:
                logger.debug("response from IAM attach_policy :" + str(response))

    def create_policy_robust(self, policy_name, policy_doc, verbose=False):
        try:
            response = self.client.create_policy(
                PolicyName=policy_name,
                PolicyDocument=policy_doc,
            )
            if verbose:
                logger.debug("response from IAM create_policy :" + str(response))
        except Exception as e:
            if 'EntityAlreadyExists' in str(e):
                try:
                    # first delete policy
                    self.remove_policy(policy_name, ignore_errors=False)
                    # recreate policy
                    response = self.client.create_policy(
                        PolicyName=policy_name,
                        PolicyDocument=policy_doc,
                    )
                    if verbose:
                        logger.debug("response from IAM create_policy :" + str(response))
                except Exception as e2:
                    raise Exception("Can't create policy %s : %s" % (policy_name, str(e2)))

    def create_instance_profile(self, verbose=False):
        try:
            self.client.create_instance_profile(
                InstanceProfileName=self.instance_profile_name
            )
        except Exception as e:
            if 'EntityAlreadyExists' in str(e):
                self.remove_instance_profile(ignore_errors=False)
                try:
                    self.client.create_instance_profile(
                        InstanceProfileName=self.instance_profile_name
                    )
                except Exception as e2:
                    raise Exception("Can't create instance profile %s: %s" % (self.instance_profile_name, str(e2)))
        # add role to instance profile
        ip = self.iam.InstanceProfile(self.instance_profile_name)
        try:
            ip.add_role(
                RoleName=self.instance_profile_name
            )
        except Exception as e:
            if 'LimitExceeded' in e:
                ip.remove_role(self.instance_profile_name, ignore_errors=False)
                try:
                    ip.add_role(
                        RoleName=self.instance_profile_name
                    )
                except Exception as e2:
                    raise Exception("Can't add role %s: %s" % (self.instance_profile_name, str(e2)))

    def create_tibanna_iam(self, verbose=False):
        """creates IAM policies and roles and a user group for tibanna
        returns prefix of all IAM policies, roles and group.
        Total 4 policies, 3 roles and 1 group is generated that is associated with a single user group
        A user group shares permission for buckets, tibanna execution and logs
        """
        # create prefix that represent a single user group
        logger.info("creating iam permissions with tibanna policy prefix %s" % self.tibanna_policy_prefix)

        # policies
        for pt in self.policy_types:
            self.create_policy_robust(self.policy_name(pt), json.dumps(self.policy_definition(pt)), verbose)

        # roles
        for rt in self.role_types:
            self.create_role_for_role_type(rt, verbose)
        # initial empty role for lambda
        self.create_empty_role_for_lambda(verbose)

        # instance profile
        # create instance profile
        self.create_instance_profile(verbose)
        # create IAM group for users who share permission
        self.create_user_group()
        return self.tibanna_policy_prefix

    def remove_role(self, rolename, verbose=False, ignore_errors=True):
        if verbose:
            logger.info("removing role %s" % rolename)
        try:
            role = self.iam.Role(rolename)
            role.description
        except Exception as e:
            if 'ResourceNotFound' in str(e) or 'NoSuchEntity' in str(e):
                if ignore_errors:
                    if verbose:
                        logger.info("role %s doesn't exist. skipping." % rolename)
                    return
                else:
                    raise Exception(e)
            raise Exception("Can't delete role %s. %s" % (rolename, str(e)))
        # first remove instance profiles attached to it
        res = self.client.list_instance_profiles_for_role(RoleName=rolename)
        for inst in res['InstanceProfiles']:
            res2 = self.client.remove_role_from_instance_profile(
                RoleName=rolename,
                InstanceProfileName=inst['InstanceProfileName']
            )
        # detach all policies
        for pol in list(role.attached_policies.all()):
            res2 = self.client.detach_role_policy(
                RoleName=rolename,
                PolicyArn=pol.arn
            )
        # delete role
        res2 = self.client.delete_role(RoleName=rolename)

    def remove_roles(self, verbose=False, ignore_errors=True):
        for rn in [self.role_name(rt) for rt in self.role_types]:
            self.remove_role(rn, verbose=verbose, ignore_errors=ignore_errors)

    def remove_instance_profile(self, verbose=False, ignore_errors=True):
        if verbose:
            logger.info("removing instance profile %s" % self.instance_profile_name)
        try:
            res = self.client.delete_instance_profile(InstanceProfileName=self.instance_profile_name)
        except Exception as e:
            if 'ResourceNotFound' in str(e) or 'NoSuchEntity' in str(e):
                if ignore_errors:
                    if verbose:
                        logger.info("instance profile %s doesn't exist. skipping." % self.instance_profile_name)
                    return
                else:
                    raise Exception(e)
            raise Exception("Can't delete instance profile. %s" % str(e))

    def remove_policy(self, policy_name, verbose=False, ignore_errors=True):
        if verbose:
            logger.info("removing policy %s" % policy_name)
        policy_arn = 'arn:aws:iam::' + self.account_id + ':policy/' + policy_name
        # first detach roles and groups and delete versions (requirements for deleting policy)
        try:
            policy = self.iam.Policy(policy_arn)
            policy.description
        except Exception as e:
            if 'ResourceNotFound' in str(e) or 'NoSuchEntity' in str(e):
                if ignore_errors:
                    if verbose:
                        logger.info("policy %s doesn't exist. skipping." % policy_arn)
                    return
                else:
                    raise Exception(e)
            raise Exception("Can't delete policy %s. %s" % (policy_arn, str(e)))
        res = self.client.list_entities_for_policy(PolicyArn=policy_arn)
        for role in res['PolicyRoles']:
            res2 = policy.detach_role(RoleName=role['RoleName'])
        for group in res['PolicyGroups']:
            res2 = policy.detach_group(GroupName=group['GroupName'])
        for v in list(policy.versions.all()):
            if not v.is_default_version:
                res2 = self.client.delete_policy_version(PolicyArn=policy_arn, VersionId=v.version_id)
        # delete policy
        res2 = self.client.delete_policy(PolicyArn=policy_arn)

    def remove_policies(self, verbose=False, ignore_errors=True):
        for pn in [self.policy_name(pt) for pt in self.policy_types]:
            self.remove_policy(pn, verbose=verbose, ignore_errors=ignore_errors)

    def detach_policies_from_group(self, verbose=False):
        try:
            # do not actually delete the group, just detach existing policies.
            # deleting a group would require users to be detached from the group.
            for pol in list(self.iam.Group(self.iam_group_name).attached_policies.all()):
                res = self.client.detach_group_policy(GroupName=self.iam_group_name, PolicyArn=pol.arn)
        except Exception as e2:
            raise Exception("Can't detach policies from group %s : %s" % (self.iam_group_name, str(e2)))

    def remove_users_from_group(self, verbose=False):
        gr = self.iam.Group(self.iam_group_name)
        for u in gr.users.iterator():
            gr.remove_user(UserName=u.user_name)

    def delete_group(self, verbose=False, ignore_errors=True):
        if verbose:
            logger.info("removing group %s" % self.iam_group_name)
        try:
            gr = self.iam.Group(self.iam_group_name)
            gr.group_id
        except Exception as e:
            if 'ResourceNotFound' in str(e) or 'NoSuchEntity' in str(e):
                if ignore_errors:
                    if verbose:
                        logger.info("group %s doesn't exist. skipping." % self.iam_group_name)
                    return
                else:
                    raise Exception(e)
            raise Exception("Can't delete policy %s. %s" % (self.iam_group_name, str(e)))

        self.remove_users_from_group(verbose)
        self.detach_policies_from_group(verbose)
        gr.delete()

    def delete_tibanna_iam(self, verbose=True, ignore_errors=True):
        self.remove_policies(verbose=verbose, ignore_errors=ignore_errors)
        self.remove_roles(verbose=verbose, ignore_errors=ignore_errors)
        self.remove_instance_profile(verbose=verbose, ignore_errors=ignore_errors)
        res = self.delete_group(verbose=verbose, ignore_errors=ignore_errors)



import boto3
import json
from datetime import datetime
from . import create_logger
from tibanna import dd_utils
from .vars import (
    STEP_FUNCTION_ARN,
    EXECUTION_ARN,
    AWS_REGION,
    DYNAMODB_TABLE
)


# logger
logger = create_logger(__name__)


class Jobs(object):

    @staticmethod
    def status(job_ids=None, exec_arns=None):
        res = dict()
        statuses = dict()
        if job_ids: 
            for j in job_ids:
                statuses.update({j: Job(job_id=j).check_status()})
        if exec_arns:
            for arn in exec_arns:
                statuses.update({arn: Job(exec_arn=arn).check_status()})
        res['completed_jobs'] = [job for job, status in iter(statuses.items()) if status == 'SUCCEEDED']
        res['failed_jobs'] = [job for job, status in iter(statuses.items()) if status == 'FAILED']
        res['running_jobs'] = [job for job, status in iter(statuses.items()) if status == 'RUNNING']
        return res


class Job(object):

    def __init__(self, job_id=None, exec_arn=None, sfn=None):
        """A job can be identified with either a job_id or an exec_arn.
        For old tibanna versions (with no dynamoDB deployed),
        a job can be identified with a job_id and sfn.
        """
        if not job_id and not exec_arn:
            raise Exception("Provide either through job id or execution arn to retrieve a job.")
        self._job_id = job_id
        self._exec_arn = exec_arn
        self.sfn = sfn  # only for old tibanna
        self._exec_desc = None
        self._log_bucket = None
        self.costupdater_exec_arn = None
        self.costupdater_exec_desc = None

        # cache for boto3 client for step function
        self._client_sfn = None

    @property
    def client_sfn(self):
        if not self._client_sfn:
            self._client_sfn = boto3.client('sfn', region=AWS_REGION)
        return self._client_sfn

    def check_costupdater_status(self):
        self.update_costupdater_exec_desc()
        return self.costupdater_exec_desc['status']

    def check_status(self):
        '''checking status of an execution.
        It works only if the execution info is still in the step function.'''
        return self.exec_desc['status']

    def check_output(self):
        '''checking status of an execution first and if it's success, get output.
        It works only if the execution info is still in the step function.'''
        if self.check_status() == 'SUCCEEDED':
            if 'output' in self.exec_desc:
                return json.loads(self.exec_desc['output'])
            else:
                return None

    @property
    def exec_arn(self):
        if self._exec_arn:
            return self._exec_arn
        elif self.job_id:  # figure out exec_arn frmo job_id
            try:
                self._exec_arn = self.get_exec_arn_from_job_id(self.job_id)
            except Exception as e:
                if self.sfn:
                    self._exec_arn = self.get_exec_arn_from_job_id_and_sfn_wo_dd(self.job_id, sfn=self.sfn)
                else:
                    raise e
            return self._exec_arn

    @property
    def exec_desc(self):
        """sfn is needed only for old tibanna
        """
        if not self._exec_desc:
            self._exec_desc = self.describe_exec(self.exec_arn)
        return self._exec_desc

    @property
    def job_id(self):
        if not self._job_id:
            if self.exec_arn:
                self._job_id = self.get_job_id_from_exec_arn(self.exec_arn)
            else:
                raise("Can't find job_id - either provide job_id or exec_arn")
        return self._job_id

    @property
    def log_bucket(self):
        if not self._log_bucket:
            if self.job_id:
                try:
                    # first try dynanmodb to get logbucket
                    self._log_bucket = self.get_log_bucket_from_job_id(self.job_id)
                except Exception as e:
                    if self.sfn:
                        self._log_bucket = self.get_log_bucket_from_job_id_and_sfn_wo_dd(self.job_id, self.sfn)
            if not self._log_bucket:
                raise Exception("Cannot retrieve log bucket.")
        return self._log_bucket

    @classmethod
    def get_log_bucket_from_job_id(cls, job_id):
        return cls.info(job_id)['Log Bucket']

    @staticmethod
    def get_log_bucket_from_job_id_and_sfn_wo_dd(job_id, sfn):
        stateMachineArn = STEP_FUNCTION_ARN(sfn)
        sf = boto3.client('stepfunctions')
        res = sf.list_executions(stateMachineArn=stateMachineArn)
        while True:
            if 'executions' not in res or not res['executions']:
                break
            breakwhile = False
            for exc in res['executions']:
                desc = sf.describe_execution(executionArn=exc['executionArn'])
                if job_id == str(json.loads(desc['input'])['jobid']):
                    logbucket = str(json.loads(desc['input'])['config']['log_bucket'])
                    breakwhile = True
                    break
            if breakwhile:
                break
            if 'nextToken' in res:
                res = sf.list_executions(nextToken=res['nextToken'],
                                         stateMachineArn=stateMachineArn)
            else:
                break
        if logbucket:
            return logbucket
        else:
            raise Exception("Cannot retrieve log bucket.")

    @staticmethod
    def get_job_id_from_exec_arn(exec_arn):
        sf = boto3.client('stepfunctions')
        desc = sf.describe_execution(executionArn=exec_arn)
        return str(json.loads(desc['input'])['jobid'])

    def update_costupdater_exec_desc(self):
        if not self.costupdater_exec_desc:
            self.update_costupdater_exec_arn_from_job_id()
            self.costupdater_exec_desc = self.describe_exec(self.costupdater_exec_arn)

    def update_costupdater_exec_arn_from_job_id(self):
        if not self.costupdater_exec_arn:
            self.costupdater_exec_arn = self.get_costupdater_exec_arn_from_job_id(self.job_id)
        # older tibanna does not have cost updater so we don't need to try the old way of doing it without dd.

    @staticmethod
    def stepfunction_exists(sfn_name):
        sf = boto3.client('stepfunctions')
        try:
            sf.describe_state_machine(stateMachineArn=STEP_FUNCTION_ARN(sfn_name))
            return True
        except Exception as e:
            if "State Machine Does Not Exist" in str(e):
                return False

    @classmethod
    def get_costupdater_exec_arn_from_job_id(cls, job_id):
        ddinfo = cls.info(job_id)
        if not ddinfo:
            raise Exception("Can't find dynamoDB entry for job_id %s" % job_id)
        exec_name = ddinfo.get('Execution Name', '')
        if not exec_name:
            raise Exception("Can't find exec_name from dynamoDB for job_id %s" % job_id)
        sfn = ddinfo.get('Step Function', '')
        if not cls.stepfunction_exists(sfn + "_costupdater"):
            raise Exception("Costupdater step function does not exist." + 
                            "To use costupdater, upgrade your tibanna to >=1.1 and redeploy with -C option!")
        exec_arn = EXECUTION_ARN(exec_name, sfn + "_costupdater")
        return exec_arn

    @classmethod
    def get_exec_arn_from_job_id(cls, job_id):
        ddinfo = cls.info(job_id)
        if not ddinfo:
            raise Exception("Can't find dynamoDB entry for job_id %s" % job_id)
        exec_name = ddinfo.get('Execution Name', '')
        if not exec_name:
            raise Exception("Can't find exec_name from dynamoDB for job_id %s" % job_id)
        sfn = ddinfo.get('Step Function', '')
        exec_arn = EXECUTION_ARN(exec_name, sfn)
        return exec_arn

    @staticmethod
    def get_exec_arn_from_job_id_and_sfn_wo_dd(job_id, sfn):
        """This is for old tibanna that did not use dyndmoDB.
        We're keeping it just for backward compatibility.
        Basically searching through all executions in a given sfn to
        find the execution that matches the job ID.
        Very slow.
        """
        stateMachineArn = STEP_FUNCTION_ARN(sfn)
        try:
            sf = boto3.client("stepfuntions", region_name=AWS_REGION)
            res = sf.list_executions(stateMachineArn=stateMachineArn)
            while True:
                if 'executions' not in res or not res['executions']:
                    break
                breakwhile = False
                for exc in res['executions']:
                    desc = sf.describe_execution(executionArn=exc['executionArn'])
                    if job_id == str(json.loads(desc['input'])['jobid']):
                        return exc['executionArn']
                if 'nextToken' in res:
                    res = sf.list_executions(nextToken=res['nextToken'],
                                             stateMachineArn=stateMachineArn)
                else:
                    break
        except:
            raise Exception("Cannot retrieve job. Try again later.")

    @classmethod
    def info(cls, job_id):
        '''returns content from dynamodb for a given job id in a dictionary form.
        returns None if the entry does not exist in dynamoDB'''
        ddres = cls.get_dd(job_id)
        return cls.get_info_from_dd(ddres)

    @staticmethod
    def describe_exec(exec_arn):
        sf = boto3.client('stepfunctions', region_name=AWS_REGION)
        return sf.describe_execution(executionArn=exec_arn)

    @staticmethod
    def get_dd(job_id):
        '''return raw content from dynamodb for a given job id'''
        ddres = dict()
        try:
            dd = boto3.client('dynamodb')
            ddres = dd.query(TableName=DYNAMODB_TABLE,
                             KeyConditions={'Job Id': {'AttributeValueList': [{'S': job_id}],
                                                       'ComparisonOperator': 'EQ'}})
            return ddres
        except Exception as e:
            logger.warning("DynamoDB entry not found: %s" % e)
            return None

    @staticmethod
    def get_info_from_dd(ddres):
        '''converts raw content from dynamodb to a dictionary form'''
        if not ddres:
            return None
        if 'Items' in ddres:
            try:
                dditem = ddres['Items'][0]
                return dd_utils.item2dict(dditem)
            except Exception as e:
                logger.warning("DynamoDB fields not found: %s" % e)
                return None
        else:
            logger.warning("DynamoDB Items field not found:")
            return None

    @staticmethod
    def add_to_dd(job_id, execution_name, sfn, logbucket, verbose=True):
        time_stamp = datetime.strftime(datetime.utcnow(), '%Y%m%d-%H:%M:%S-UTC')
        dydb = boto3.client('dynamodb', region_name=AWS_REGION)
        try:
            # first check the table exists
            res = dydb.describe_table(TableName=DYNAMODB_TABLE)
        except Exception as e:
            if verbose:
                logger.info("Not adding to dynamo table: %s" % e)
            return
        try:
            response = dydb.put_item(
                TableName=DYNAMODB_TABLE,
                Item={
                    'Job Id': {
                        'S': job_id
                    },
                    'Execution Name': {
                        'S': execution_name
                    },
                    'Step Function': {
                        'S': sfn
                    },
                    'Log Bucket': {
                        'S': logbucket
                    },
                    'Time Stamp': {
                        'S': time_stamp
                    }
                }
            )
            if verbose:
                logger.info("Successfully put item to dynamoDB: " + str(response))
        except Exception as e:
            raise(e)




import datetime


class Top(object):
    """class TopSeries stores the information of a series of top commands

    ::
        echo -n 'Timestamp: '; date +%F-%H:%M:%S
        top -b -n1 [-i] [-c]

    over short intervals to monitor the same set of processes over time. 

    An example input content looks like below, or a series of these.
    The initialization works at any time interval and can be used as a generic
    class, but the class is designed for the output of a regular top commands above
    run at about 1-minute intervals, which is performed by awsf3 on an AWSEM instance
    through cron jobs. (some can be skipped but there should be no more than 1 per minute).
    This top output can be obtained through ``tibanna log -j <job_id> -t`` or through
    API ``API().log(job_id=<job_id>, top=True)``.

    ::

        Timestamp: 2020-12-18-18:55:37
        top - 18:55:37 up 4 days,  2:37,  0 users,  load average: 5.59, 5.28, 5.76
        Tasks:   7 total,   1 running,   6 sleeping,   0 stopped,   0 zombie
        %Cpu(s):  6.6 us,  0.1 sy,  0.0 ni, 93.2 id,  0.0 wa,  0.0 hi,  0.0 si,  0.0 st
        KiB Mem : 12971188+total, 10379019+free, 20613644 used,  5308056 buff/cache
        KiB Swap:        0 total,        0 free,        0 used. 10834606+avail Mem 

          PID USER      PR  NI    VIRT    RES    SHR S  %CPU %MEM     TIME+ COMMAND
          712 root      20   0 36.464g 8.223g  19572 S 100.0  6.6 125:55.12 java -Xmx32g -Xms32g -jar juicer_tools.jar addNorm -w 1000 -d -F out.hic
          17919 ubuntu    20   0   40676   3828   3144 R   6.2  0.0   0:00.01 top -b -n1 -c -i -w 10000

    The default timestamp from top output does not contain dates, which can screw up multi-day processes
    which is common for bioinformatics pipelines. So, an extra timestamp is added before each top command.

    To parse top output content, simply create an object. This will create processes attribute,
    which is a raw parsed result organized by time stamps.

    ::

        top = Top(top_output_content)

    To reorganize the contents by commands, run digest. By default, the max number of commands is 16,
    and if there are more than 16 unique commands, they will be collapsed into prefixes.

    ::

        top.digest()

    To write a csv / tsv file organized by both timestamps (rows) and commands (columns),
    use :func: write_to_csv.

    ::

        top.write_to_csv(...)

    """

    # assume this format for timestamp
    timestamp_format = '%Y-%m-%d-%H:%M:%S'

    # These commands are excluded when parsing the top output
    # Currently only 1-, 2- or 3-word prefixes work.
    exclude_list = ['top', 'docker', 'dockerd', '/usr/bin/dockerd', 'cron',
                    'docker-untar', 'containerd', 'goofys-latest', 'cwltool',
                    '/usr/bin/containerd-shim-runc-v2', 'goofys', 'nodejs --eval',
                    '/usr/bin/python3 /usr/local/bin/cwltool', 'containerd-shim',
                    '/usr/bin/python3 /bin/unattended-upgrade',
                    '/usr/bin/python3 /usr/local/bin/awsf3',
                    '/usr/bin/python3 /usr/local/bin/aws s3',
                    'java -jar /usr/local/bin/cromwell.jar',
                    'java -jar /usr/local/bin/cromwell-35.jar']

    def __init__(self, contents):
        """initialization parsed top output content and
        creates processes which is a dictionary with timestamps as keys
        and a list of Process class objects as a value.
        It also creates empty attributes timestamps, commands, cpus and mems
        which can be filled through method :func: digest.
        """
        self.processes = dict()
        self.timestamps = []
        self.commands = []
        self.cpus = dict()
        self.mems = dict()
        self.parse_contents(contents)

    def parse_contents(self, contents):
        is_in_table = False
        for line in contents.splitlines():
            if line.startswith('Timestamp:'):
                timestamp = line.split()[1]
                continue
            if line.lstrip().startswith('PID'):
                is_in_table = True
                continue
            if not line or line.isspace():
                is_in_table = False
            if is_in_table:
                if timestamp not in self.processes:
                    self.processes[timestamp] = []
                process = Process(line)
                if not self.should_skip_process(process):
                    self.processes[timestamp].append(Process(line))

    def digest(self, max_n_commands=16, sort_by='alphabetical'):
        """Fills in timestamps, commands, cpus and mems attributes
        from processes attribute.
        :param max_n_commands: When the number of unique commands exceeds
        this value, they are collapsed into unique prefixes.
        :sort_by: alphabetical|cpu|mem The commands are by default sorted
        alphabetically, but optionally can be sorted by total cpus or total
        mem (in reverser order) (e.g. the first command consumed the most cpu)
        """
        # Reinitializat these so that you get the same results if you run it twice
        self.timestamps = []
        self.commands = []
        self.cpus = dict()
        self.mems = dict()
        # First fill in commands from commands in processes (and collapse if needed.)
        self.commands = self.get_collapsed_commands(max_n_commands)
        # Fill in timestamps, cpus and mems from processes, matching collapsed commands.
        self.nTimepoints = len(self.processes)
        timestamp_ind = 0
        for timestamp in sorted(self.processes):
            # sorted timestamps (columns)
            self.timestamps.append(timestamp)
            # commands (rows)
            for process in self.processes[timestamp]:
                # find a matching collapsed command (i.e. command prefix) and use that as command.
                command = Top.convert_command_to_collapsed_command(process.command, self.commands)
                if command not in self.cpus:
                    self.cpus[command] = [0] * self.nTimepoints
                    self.mems[command] = [0] * self.nTimepoints
                self.cpus[command][timestamp_ind] += process.cpu
                self.mems[command][timestamp_ind] += process.mem
            timestamp_ind += 1
        # sort commands according to total cpu
        self.sort_commands(by=sort_by)

    def get_collapsed_commands(self, max_n_commands):
        """If the number of commands exceeds max_n_commands,
        return a collapsed set of commands
        that consists of prefixes of commands so that
        the total number is within max_n_commands.
        First decide the number of words from the beginning of the commands
        to collapse commands that start with the same words, i.e.
        find the maximum number of words that makes the number of unique commands to be
        bounded by max_n_commands.
        If using only the first word is not sufficient, go down to the characters of
        the first word. If that's still not sufficient, collapse all of them into a single
        command ('all_commands')
        After the collapse, commands that are unique to a collapsed prefix are
        extended back to the original command.
        """

        all_commands = set()
        for timestamp in self.processes:
            all_commands.update(set([pr.command for pr in self.processes[timestamp]]))

        if len(all_commands) <= max_n_commands:
            # no need to collapse
            return list(all_commands)

        # decide the number of words from the beginning of the commands
        # to collapse commands starting with the same words
        all_cmd_lengths = [len(cmd.split()) for cmd in all_commands]  # number of words per command
        max_cmd_length = max(all_cmd_lengths)
        min_cmd_length = min(all_cmd_lengths)
        collapsed_len = max_cmd_length - 1
        n_commands = len(all_commands)
        while(n_commands > max_n_commands and collapsed_len > 1):
            reduced_commands = set()
            for cmd in all_commands:
                reduced_commands.add(Top.first_words(cmd, collapsed_len))
            n_commands = len(reduced_commands)
            collapsed_len -= 1

        # went down to the first words but still too many commands - start splitting characters then
        if n_commands > max_n_commands:
            all_cmd_lengths = [len(cmd.split()[0]) for cmd in all_commands]  # number of characters of the first word
            max_cmd_length = max(all_cmd_lengths)
            min_cmd_length = min(all_cmd_lengths)
            collapsed_len = max_cmd_length - 1
            while(n_commands > max_n_commands and collapsed_len > 1):
                reduced_commands = set()
                for cmd in all_commands:
                    reduced_commands.add(Top.first_characters(cmd.split()[0], collapsed_len))
                n_commands = len(reduced_commands)
                collapsed_len -= 1

        if n_commands > max_n_commands:
            return ['all_commands']
        else:
            # extend reduced commands that don't need to be reduced
            for r_cmd in list(reduced_commands):  # wrap in list so that we can remove elements in the loop
                uniq_cmds = [cmd for cmd in all_commands if cmd.startswith(r_cmd)]
                if len(uniq_cmds) == 1:
                    reduced_commands.remove(r_cmd)
                    reduced_commands.add(uniq_cmds[0])
            return reduced_commands

    def write_to_csv(self, csv_file, metric='cpu', delimiter=',', colname_for_timestamps='timepoints',
                     timestamp_start=None, timestamp_end=None, base=0):
        """write metrics as csv file with commands as columns
        :param metric: 'cpu' or 'mem'
        :param delimiter: default ','
        :param colname_for_timestamps: colunm name for the timepoint column (1st column). default 'timepoints'
        :param timestamp_start: start time in the same timestamp format (e.g. 01:23:45),
                                time stamps will be converted to minutes since start time.
                                The report starts with minute 0.
                                Time points with no top records will be filled with 0.
                                If not specified, the first timestamp in the top commands will be used.
        :param timestamp_end: end time in the same timestamp format (e.g. 01:23:45),
                              The reports will be generated only up to the end time.
                              Time points with no top records will be filled with 0.
                              If not specified, the last timestamp in the top commands will be used.
        :param base: default 0. If 0, minutes start with 0, if 1, minutes are 1-based (shifted by 1).
        """
        metric_array = getattr(self, metric + 's')
        if self.timestamps:
            if not timestamp_start:
                timestamp_start = self.timestamps[0]
            if not timestamp_end:
                timestamp_end = self.timestamps[-1]
            timestamps_as_minutes = self.timestamps_as_minutes(timestamp_start)
            last_minute = self.as_minutes(timestamp_end, timestamp_start)
        else:  # default when timestamps is not available (empty object)
            timestamps_as_minutes = range(0, 5)
            last_minute = 5
        with open(csv_file, 'w') as fo:
            # header
            # we have to escape any double quotes that are present in the cmd, before wrapping it in double quotes. Otherwise we 
            # will get incorrect column counts when creating the metrics report.
            fo.write(delimiter.join([colname_for_timestamps] + [Top.wrap_in_double_quotes(cmd.replace('"', '""')) for cmd in self.commands]))
            fo.write('\n')
            # contents
            # skip timepoints earlier than timestamp_start
            for i in range(0, len(timestamps_as_minutes)):
                if timestamps_as_minutes[i] >= 0:
                    break
            for clock in range(0, last_minute + 1):
                clock_shifted = clock + base
                if i < len(timestamps_as_minutes) and timestamps_as_minutes[i] == clock:
                    fo.write(delimiter.join([str(clock_shifted)] + [str(metric_array[cmd][i]) for cmd in self.commands]))
                    i += 1
                else:
                    fo.write(delimiter.join([str(clock_shifted)] + ['0' for cmd in self.commands]))  # add 0 for timepoints not reported
                fo.write('\n')

    def should_skip_process(self, process):
        """A predicate function to check if the process should be skipped (excluded).
        It returns True if the input process should be skipped.
        e.g. the top command itself is excluded, as well as docker, awsf3, cwltool, etc.
        the list to be excluded is in self.exclude_list.
        It compares either first word or first two or three words only.
        Kernel threads (single-word commands wrapped in bracket (e.g. [perl]) are also excluded.
        """
        first_word = Top.first_words(process.command, 1)
        first_two_words = Top.first_words(process.command, 2)
        first_three_words = Top.first_words(process.command, 3)
        if first_word in self.exclude_list:
            return True
        elif first_two_words in self.exclude_list:
            return True
        elif first_three_words in self.exclude_list:
            return True
        if first_word.startswith('[') and first_word.endswith(']'):
            return True
        return False

    @staticmethod
    def convert_command_to_collapsed_command(cmd, collapsed_commands):
        if collapsed_commands == 'all_commands':  # collapsed to one command
            return 'all_commands'
        elif cmd in collapsed_commands:  # not collapsed
            return cmd
        else:  # collapsed to prefix
            all_prefixes = [_ for _ in collapsed_commands if cmd.startswith(_)]
            longest_prefix = sorted(all_prefixes, key=lambda x: len(x), reverse=True)[0]
            return longest_prefix

    def total_cpu_per_command(self, command):
        return sum([v for v in self.cpus[command]])

    def total_mem_per_command(self, command):
        return sum([v for v in self.mems[command]])

    def sort_commands(self, by='cpu'):
        """sort self.commands by total cpu (default) or mem in reverse order,
           or alphabetically (by='alphabetical')"""
        if by == 'cpu':
            self.commands = sorted(self.commands, key=lambda x: self.total_cpu_per_command(x), reverse=True)
        elif by == 'mem':
            self.commands = sorted(self.commands, key=lambda x: self.total_mem_per_command(x), reverse=True)
        elif by == 'alphabetical':
            self.commands = sorted(self.commands)

    @classmethod
    def as_minutes(cls, timestamp, timestamp_start):
        """timestamp as minutes since timestamp_start.
        :param timestamp: given timestamp in the same format (e.g. 01:23:45)
        :param timestamp_start: start timestamp in the same format (e.g. 01:20:45)
        In the above example, 3 will be the return value.
        """ 
        dt = cls.as_datetime(timestamp)
        dt_start = cls.as_datetime(timestamp_start)
        # negative numbers are not supported by timedelta, so do each case separately
        if dt > dt_start:
            return round((dt - dt_start).seconds / 60)
        else:
            return -round((dt_start - dt).seconds / 60)

    def timestamps_as_minutes(self, timestamp_start):
        """convert self.timestamps to a list of minutes since timestamp_start
        :param timestamp_start: timestamp in the same format (e.g. 01:23:45)
        """
        return [self.as_minutes(t, timestamp_start) for t in self.timestamps]

    @classmethod
    def as_datetime(cls, timestamp):
        return datetime.datetime.strptime(timestamp, cls.timestamp_format)

    @staticmethod
    def wrap_in_double_quotes(string):
        """wrap a given string with double quotes (e.g. haha -> "haha")
        """
        return '\"' + string + '\"'

    @staticmethod
    def first_words(string, n_words):
        """returns first n words of a string
        e.g. first_words('abc def ghi', 2) ==> 'abc def'
        """
        words = string.split()
        return ' '.join(words[0:min(n_words, len(words))])

    @staticmethod
    def first_characters(string, n_letters):
        """returns first n letters of a string
        e.g. first_characters('abc def ghi', 2) ==> 'ab'
        """
        letters = list(string)
        return ''.join(letters[0:min(n_letters, len(letters))])

    def as_dict(self):
        return self.__dict__


class Process(object):             
    def __init__(self, top_line):
        prinfo_as_list = top_line.lstrip().split()
        self.pid = prinfo_as_list[0]
        self.user = prinfo_as_list[1]
        self.cpu = float(prinfo_as_list[8])
        self.mem = float(prinfo_as_list[9])
        self.command = ' '.join(prinfo_as_list[11:])

    def as_dict(self):
        return self.__dict__



#!/usr/bin/python
import json
import time
import os
import logging
import boto3
import botocore
import re
from . import create_logger
from datetime import datetime, timedelta
from .utils import (
    does_key_exist,
    read_s3,
    put_object_s3
)
from .vars import (
    AWS_REGION,
    AWS_REGION_NAMES
)
from .exceptions import (
    PricingRetrievalException
)


logger = create_logger(__name__)


def get_cost(postrunjson, job_id):

    job = postrunjson.Job

    def reformat_time(t, delta):
        d = datetime.strptime(t, '%Y%m%d-%H:%M:%S-UTC') + timedelta(days=delta)
        return d.strftime("%Y-%m-%d")

    start_time = reformat_time(job.start_time, -1)  # give more room
    if(job.end_time != None):
        end_time = reformat_time(job.end_time, 1)
    else:
        end_time = datetime.utcnow() + timedelta(days=1) # give more room
        end_time = end_time.strftime("%Y-%m-%d")

    billing_args = {'Filter': {'Tags': {'Key': 'Name', 'Values': ['awsem-' + job_id]}},
                    'Granularity': 'DAILY',
                    'TimePeriod': {'Start': start_time,
                                    'End': end_time},
                    'Metrics': ['BlendedCost'],
                    }

    try:
        billingres = boto3.client('ce').get_cost_and_usage(**billing_args)
    except botocore.exceptions.ClientError as e:
        logger.warning("%s. Please try to deploy the latest version of Tibanna." % e)
        return 0.0

    cost = sum([float(_['Total']['BlendedCost']['Amount']) for _ in billingres['ResultsByTime']])
    return cost


def get_cost_estimate(postrunjson, ebs_root_type = "gp3", aws_price_overwrite = None):
    """
    aws_price_overwrite can be used to overwrite the prices obtained from AWS (e.g. ec2 spot price).
    This allows historical cost estimates. It is also used for testing. It is a dictionary with keys:
    ec2_spot_price, ec2_ondemand_price, ebs_root_storage_price, ebs_storage_price,
    ebs_iops_price (gp3, io1), ebs_io2_iops_prices, ebs_throughput_price
    """

    cfg = postrunjson.config
    job = postrunjson.Job
    estimated_cost = 0.0

    if(job.end_time == None):
        logger.warning("job.end_time not available. Cannot calculate estimated cost.")
        return 0.0, "NA"

    job_start = datetime.strptime(job.start_time, '%Y%m%d-%H:%M:%S-UTC')
    job_end = datetime.strptime(job.end_time, '%Y%m%d-%H:%M:%S-UTC')
    job_duration = (job_end - job_start).seconds / 3600.0 # in hours

    try:
        pricing_client = boto3.client('pricing', region_name=AWS_REGION)

        # Get EC2 spot price
        if(cfg.spot_instance):
            if(cfg.spot_duration):
                raise PricingRetrievalException("Pricing with spot_duration is not supported")

            if(not job.instance_availablity_zone):
                raise PricingRetrievalException("Instance availability zone is not available. You might have to deploy a newer version of Tibanna.")

            ec2_client=boto3.client('ec2',region_name=AWS_REGION)
            prices=ec2_client.describe_spot_price_history(
                InstanceTypes=[cfg.instance_type],
                ProductDescriptions=['Linux/UNIX'],
                AvailabilityZone=job.instance_availablity_zone,
                MaxResults=1) # Most recent price is on top

            if(len(prices['SpotPriceHistory']) == 0):
                raise PricingRetrievalException("Spot price could not be retrieved")

            ec2_spot_price = (float)(prices['SpotPriceHistory'][0]['SpotPrice'])

            if((aws_price_overwrite is not None) and 'ec2_spot_price' in aws_price_overwrite):
                ec2_spot_price = aws_price_overwrite['ec2_spot_price']

            estimated_cost = estimated_cost + ec2_spot_price * job_duration

        else: # EC2 onDemand Prices

            prices = pricing_client.get_products(ServiceCode='AmazonEC2', Filters=[
                {
                    'Type': 'TERM_MATCH',
                    'Field': 'instanceType',
                    'Value': cfg.instance_type
                },
                {
                    'Type': 'TERM_MATCH',
                    'Field': 'operatingSystem',
                    'Value': 'Linux'
                },
                {
                    'Type': 'TERM_MATCH',
                    'Field': 'location',
                    'Value': AWS_REGION_NAMES[AWS_REGION]
                },
                {
                    'Type': 'TERM_MATCH',
                    'Field': 'preInstalledSw',
                    'Value': 'NA'
                },
                {
                    'Type': 'TERM_MATCH',
                    'Field': 'capacitystatus',
                    'Value': 'used'
                },
                {
                    'Type': 'TERM_MATCH',
                    'Field': 'tenancy',
                    'Value': 'Shared'
                },
            ])
            price_list = prices["PriceList"]

            if(not prices["PriceList"] or len(price_list) == 0):
                raise PricingRetrievalException("We could not retrieve EC2 prices from Amazon")

            if(len(price_list) > 1):
                raise PricingRetrievalException("EC2 prices are ambiguous")

            price_item = json.loads(price_list[0])
            terms = price_item["terms"]
            term = list(terms["OnDemand"].values())[0]
            price_dimension = list(term["priceDimensions"].values())[0]
            ec2_ondemand_price = (float)(price_dimension['pricePerUnit']["USD"])


            if((aws_price_overwrite is not None) and 'ec2_ondemand_price' in aws_price_overwrite):
                ec2_ondemand_price = aws_price_overwrite['ec2_ondemand_price']

            estimated_cost = estimated_cost + ec2_ondemand_price * job_duration


        # Get EBS pricing

        prices = pricing_client.get_products(ServiceCode='AmazonEC2', Filters=[
            {
                'Type': 'TERM_MATCH',
                'Field': 'location',
                'Value': AWS_REGION_NAMES[AWS_REGION]
            },
            {
                'Field': 'volumeApiName',
                'Type': 'TERM_MATCH',
                'Value': ebs_root_type,
            },
            {
                'Field': 'productFamily',
                'Type': 'TERM_MATCH',
                'Value': 'Storage',
            },
        ])
        price_list = prices["PriceList"]

        if(not prices["PriceList"] or len(price_list) == 0):
            raise PricingRetrievalException("We could not retrieve EBS prices from Amazon")

        if(len(price_list) > 1):
            raise PricingRetrievalException("EBS prices are ambiguous")

        price_item = json.loads(price_list[0])
        terms = price_item["terms"]
        term = list(terms["OnDemand"].values())[0]
        price_dimension = list(term["priceDimensions"].values())[0]
        ebs_root_storage_price = (float)(price_dimension['pricePerUnit']["USD"])

        if((aws_price_overwrite is not None) and 'ebs_root_storage_price' in aws_price_overwrite):
            ebs_root_storage_price = aws_price_overwrite['ebs_root_storage_price']

        # add root EBS costs
        root_ebs_cost = ebs_root_storage_price * cfg.root_ebs_size * job_duration / (24.0*30.0)
        estimated_cost = estimated_cost + root_ebs_cost

        # add additional EBS costs
        if(cfg.ebs_type == "gp3"):
            ebs_storage_cost = ebs_root_storage_price * cfg.ebs_size * job_duration / (24.0*30.0)
            estimated_cost = estimated_cost + ebs_storage_cost

            # Add throughput
            if(cfg.ebs_throughput):
                prices = pricing_client.get_products(ServiceCode='AmazonEC2', Filters=[
                    {
                        'Type': 'TERM_MATCH',
                        'Field': 'location',
                        'Value': AWS_REGION_NAMES[AWS_REGION]
                    },
                    {
                        'Field': 'volumeApiName',
                        'Type': 'TERM_MATCH',
                        'Value': cfg.ebs_type,
                    },
                    {
                        'Field': 'productFamily',
                        'Type': 'TERM_MATCH',
                        'Value': 'Provisioned Throughput',
                    },
                ])
                price_list = prices["PriceList"]

                if(not prices["PriceList"] or len(price_list) == 0):
                    raise PricingRetrievalException("We could not retrieve EBS throughput prices from Amazon")

                if(len(price_list) > 1):
                    raise PricingRetrievalException("EBS throughput prices are ambiguous")

                price_item = json.loads(price_list[0])
                terms = price_item["terms"]
                term = list(terms["OnDemand"].values())[0]
                price_dimension = list(term["priceDimensions"].values())[0]
                ebs_throughput_price = (float)(price_dimension['pricePerUnit']["USD"])/1000 # unit: mbps

                if((aws_price_overwrite is not None) and 'ebs_throughput_price' in aws_price_overwrite):
                    ebs_throughput_price = aws_price_overwrite['ebs_throughput_price']

                free_tier = 125
                ebs_throughput_cost = ebs_throughput_price * max(cfg.ebs_throughput - free_tier, 0) * job_duration / (24.0*30.0)
                estimated_cost = estimated_cost + ebs_throughput_cost

        else:
            prices = pricing_client.get_products(ServiceCode='AmazonEC2', Filters=[
                {
                    'Type': 'TERM_MATCH',
                    'Field': 'location',
                    'Value': AWS_REGION_NAMES[AWS_REGION]
                },
                {
                    'Field': 'volumeApiName',
                    'Type': 'TERM_MATCH',
                    'Value': cfg.ebs_type,
                },
                {
                    'Field': 'productFamily',
                    'Type': 'TERM_MATCH',
                    'Value': 'Storage',
                },
            ])
            price_list = prices["PriceList"]

            if(not prices["PriceList"] or len(price_list) == 0):
                raise PricingRetrievalException("We could not retrieve EBS prices from Amazon")

            if(len(price_list) > 1):
                raise PricingRetrievalException("EBS prices are ambiguous")

            price_item = json.loads(price_list[0])
            terms = price_item["terms"]
            term = list(terms["OnDemand"].values())[0]
            price_dimension = list(term["priceDimensions"].values())[0]
            ebs_storage_price = (float)(price_dimension['pricePerUnit']["USD"])

            if((aws_price_overwrite is not None) and 'ebs_storage_price' in aws_price_overwrite):
                ebs_storage_price = aws_price_overwrite['ebs_storage_price']

            add_ebs_cost = ebs_storage_price * cfg.ebs_size * job_duration / (24.0*30.0)
            estimated_cost = estimated_cost + add_ebs_cost

        ## IOPS PRICING
        # Add IOPS prices for io1 or gp3
        if( (cfg.ebs_type == "io1" or cfg.ebs_type == "gp3") and cfg.ebs_iops):
            prices = pricing_client.get_products(ServiceCode='AmazonEC2', Filters=[
                {
                    'Type': 'TERM_MATCH',
                    'Field': 'location',
                    'Value': AWS_REGION_NAMES[AWS_REGION]
                },
                {
                    'Field': 'volumeApiName',
                    'Type': 'TERM_MATCH',
                    'Value': cfg.ebs_type,
                },
                {
                    'Field': 'productFamily',
                    'Type': 'TERM_MATCH',
                    'Value': 'System Operation',
                },
            ])
            price_list = prices["PriceList"]

            if(not prices["PriceList"] or len(price_list) == 0):
                raise PricingRetrievalException("We could not retrieve EBS prices from Amazon")
            if(len(price_list) > 1):
                raise PricingRetrievalException("EBS prices are ambiguous")

            price_item = json.loads(price_list[0])
            terms = price_item["terms"]
            term = list(terms["OnDemand"].values())[0]
            price_dimension = list(term["priceDimensions"].values())[0]
            ebs_iops_price = (float)(price_dimension['pricePerUnit']["USD"])

            if((aws_price_overwrite is not None) and 'ebs_iops_price' in aws_price_overwrite):
                ebs_iops_price = aws_price_overwrite['ebs_iops_price']

            if cfg.ebs_type == "gp3":
                free_tier = 3000
                ebs_iops_cost = ebs_iops_price * max(cfg.ebs_iops - free_tier, 0) * job_duration / (24.0*30.0)
            else:
                ebs_iops_cost = ebs_iops_price * cfg.ebs_iops * job_duration / (24.0*30.0)

            estimated_cost = estimated_cost + ebs_iops_cost

        elif (cfg.ebs_type == "io2" and cfg.ebs_iops):
            prices = pricing_client.get_products(ServiceCode='AmazonEC2', Filters=[
                {
                    'Type': 'TERM_MATCH',
                    'Field': 'location',
                    'Value': AWS_REGION_NAMES[AWS_REGION]
                },
                {
                    'Field': 'volumeApiName',
                    'Type': 'TERM_MATCH',
                    'Value': cfg.ebs_type,
                },
                {
                    'Field': 'productFamily',
                    'Type': 'TERM_MATCH',
                    'Value': 'System Operation',
                },
            ])
            price_list = prices["PriceList"]

            if(len(price_list) != 3):
                raise PricingRetrievalException("EBS prices for io2 are incomplete")

            ebs_io2_iops_prices = []
            for price_entry in price_list:
                price_item = json.loads(price_entry)
                terms = price_item["terms"]
                term = list(terms["OnDemand"].values())[0]
                price_dimension = list(term["priceDimensions"].values())[0]
                ebs_iops_price = (float)(price_dimension['pricePerUnit']["USD"])
                ebs_io2_iops_prices.append(ebs_iops_price)
            ebs_io2_iops_prices.sort(reverse=True)

            if((aws_price_overwrite is not None) and 'ebs_io2_iops_prices' in aws_price_overwrite):
                ebs_io2_iops_prices = aws_price_overwrite['ebs_io2_iops_prices']

            # Pricing tiers are currently hardcoded. There wasn't a simple way to extract them from the pricing information
            tier0 = 32000
            tier1 = 64000

            ebs_iops_cost = (
                ebs_io2_iops_prices[0] * min(cfg.ebs_iops, tier0) + # Portion below 32000 IOPS
                ebs_io2_iops_prices[1] * min(max(cfg.ebs_iops - tier0, 0), tier1 - tier0) + # Portion between 32001 and 64000 IOPS
                ebs_io2_iops_prices[2] * max(cfg.ebs_iops - tier1, 0) # Portion above 64000 IOPS
                ) * job_duration / (24.0*30.0)
            estimated_cost = estimated_cost + ebs_iops_cost

        time_since_run = (datetime.utcnow() - job_end).total_seconds() / (3600 * 24) # days
        estimation_type = "retrospective estimate" if time_since_run > 10 else "immediate estimate"

        return estimated_cost, estimation_type

    except botocore.exceptions.ClientError as e:
        logger.warning("Cost estimation error: %s. Please try to deploy the latest version of Tibanna." % e)
        return 0.0, "NA"
    except PricingRetrievalException as e:
        logger.warning("Cost estimation error: %s" % e)
        return 0.0, "NA"
    except Exception as e:
        logger.warning("Cost estimation error: %s" % e)
        return 0.0, "NA"


def get_cost_estimate_from_tsv(log_bucket, job_id):

    s3_key = os.path.join(job_id + '.metrics/', 'metrics_report.tsv')
    cost_estimate = 0.0
    cost_estimate_type = "NA"

    if(does_key_exist(log_bucket, s3_key) == False):
        return cost_estimate, cost_estimate_type

    try:
        read_file = read_s3(log_bucket, s3_key)
        for row in read_file.splitlines():
            line = row.split("\t")
            if(line[0] == "Estimated_Cost"):
                cost_estimate = float(line[1])
            if(line[0] == "Estimated_Cost_Type"):
                cost_estimate_type = line[1]
    except Exception as e:
        logger.warning("Could not get cost estimate from tsv: %s" % e)
        pass

    return cost_estimate, cost_estimate_type


def update_cost_estimate_in_tsv(log_bucket, job_id, cost_estimate, cost_estimate_type, encryption=False, kms_key_id=None):

    s3_key = os.path.join(job_id + '.metrics/', 'metrics_report.tsv')

    if(does_key_exist(log_bucket, s3_key) == False):
        return

    # reading from metrics_report.tsv
    read_file = read_s3(log_bucket, s3_key)

    # get the current estimate type in the file
    for row in read_file.splitlines():
        line = row.split("\t")
        if(line[0] == "Estimated_Cost_Type"):
            current_cost_estimate_type = line[1]

    if(cost_estimate_type=="retrospective estimate" and (current_cost_estimate_type=="immediate estimate" or current_cost_estimate_type=="actual cost") ):
        logger.warning("There already is a probably more accurate estimate in the tsv. Not updating.")
        return

    write_file = ""
    for row in read_file.splitlines():
        # Remove Estimated_Cost and Estimated_Cost_Type from file, since we want to update it
        if("Estimated_Cost" in row.split("\t") or "Estimated_Cost_Type" in row.split("\t")):
            continue
        if("Cost" in row.split("\t") and cost_estimate_type=="actual cost"):
            continue
        write_file = write_file + row + '\n'

    if(cost_estimate_type=="actual cost"):
        write_file = write_file + 'Cost\t' + str(cost_estimate) + '\n'
    write_file = write_file + 'Estimated_Cost\t' + str(cost_estimate) + '\n'
    write_file = write_file + 'Estimated_Cost_Type\t' + cost_estimate_type + '\n'
    put_object_s3(content=write_file, key=s3_key, bucket=log_bucket,
                  encrypt_s3_upload=encryption, kms_key_id=kms_key_id)


def update_cost_in_tsv(log_bucket, job_id, cost,
                       encryption=False, kms_key_id=None):

    s3_key = os.path.join(job_id + '.metrics/', 'metrics_report.tsv')

    if(does_key_exist(log_bucket, s3_key) == False):
        return

    # reading from metrics_report.tsv
    read_file = read_s3(log_bucket, s3_key)

    write_file = ""
    for row in read_file.splitlines():
        # Remove Cost from file, since we want to update it
        if("Cost" not in row.split("\t")):
            write_file = write_file + row + '\n'

    write_file = write_file + 'Cost\t' + str(cost) + '\n'
    put_object_s3(content=write_file, key=s3_key, bucket=log_bucket,
                  encrypt_s3_upload=encryption, kms_key_id=kms_key_id)



# -*- coding: utf-8 -*-
import os
import boto3
import botocore
import json
import time
import copy
import importlib
import shutil
import subprocess
import webbrowser
from datetime import datetime, timedelta
from dateutil.tz import tzutc
from uuid import uuid4, UUID
from types import ModuleType
from . import create_logger
from .vars import (
    _tibanna,
    AWS_ACCOUNT_NUMBER,
    AWS_REGION,
    TIBANNA_DEFAULT_STEP_FUNCTION_NAME,
    STEP_FUNCTION_ARN,
    EXECUTION_ARN,
    AMI_ID,
    TIBANNA_REPO_NAME,
    TIBANNA_REPO_BRANCH,
    TIBANNA_PROFILE_ACCESS_KEY,
    TIBANNA_PROFILE_SECRET_KEY,
    METRICS_URL,
    DYNAMODB_TABLE,
    DYNAMODB_KEYNAME,
    SFN_TYPE,
    LAMBDA_TYPE,
    RUN_TASK_LAMBDA_NAME,
    CHECK_TASK_LAMBDA_NAME,
    UPDATE_COST_LAMBDA_NAME,
    S3_ENCRYT_KEY_ID
)
from .utils import (
    _tibanna_settings,
    create_jobid,
    does_key_exist,
    read_s3,
    upload,
    put_object_s3,
    retrieve_all_keys,
    delete_keys,
    create_tibanna_suffix
)
from .ec2_utils import (
    UnicornInput,
    upload_workflow_to_s3
)
from .pricing_utils import (
    get_cost,
    get_cost_estimate,
    update_cost_estimate_in_tsv,
    update_cost_in_tsv,
    get_cost_estimate_from_tsv
)
from .job import Job
from .ami import AMI
from ._version import __version__
# from botocore.errorfactory import ExecutionAlreadyExists
from .stepfunction import StepFunctionUnicorn
from .stepfunction_cost_updater import StepFunctionCostUpdater
from .awsem import AwsemRunJson, AwsemPostRunJson
from .exceptions import (
    MetricRetrievalException
)
from . import dd_utils


# logger
logger = create_logger(__name__)


UNICORN_LAMBDAS = ['run_task_awsem', 'check_task_awsem']


class API(object):

    # This one cannot be imported in advance, because it causes circular import.
    # lambdas run_workflow / validate_md5_s3_initiator needs to import this API
    # to call run_workflow (this is a pony problem but to be consistent, I add
    # lambdas_module here for unicorn as well.
    @property
    def lambdas_module(self):
        from . import lambdas as unicorn_lambdas
        return unicorn_lambdas

    @property
    def lambda_names(self):
        lambdas = [mod for mod in dir(self.lambdas_module)
            if isinstance(getattr(self.lambdas_module, mod), ModuleType)]
        # We remove update_cost_awsem, as this typically treated separately
        lambdas.remove(self.update_cost_lambda)
        return lambdas

    @property
    def tibanna_packages(self):
        import tibanna
        return [tibanna]

    StepFunction = StepFunctionUnicorn
    StepFunctionCU = StepFunctionCostUpdater
    default_stepfunction_name = TIBANNA_DEFAULT_STEP_FUNCTION_NAME
    default_env = ''
    sfn_type = SFN_TYPE
    lambda_type = LAMBDA_TYPE

    run_task_lambda = RUN_TASK_LAMBDA_NAME
    check_task_lambda = CHECK_TASK_LAMBDA_NAME
    update_cost_lambda = UPDATE_COST_LAMBDA_NAME

    ami_id = AMI_ID

    @property
    def UNICORN_LAMBDAS(self):
        return [self.run_task_lambda, self.check_task_lambda]

    @property
    def do_not_delete(self):
        return []  # list of lambda names that should not be deleted before updating

    @property
    def TibannaResource(self):
        from .cw_utils import TibannaResource
        return TibannaResource

    @property
    def IAM(self):
        from .iam_utils import IAM
        return IAM

    def __init__(self):
        pass

    def randomize_run_name(self, run_name, sfn):
        arn = EXECUTION_ARN(run_name, sfn)
        client = boto3.client('stepfunctions', region_name=AWS_REGION)
        try:
            response = client.describe_execution(
                    executionArn=arn
            )
            if response:
                if len(run_name) > 36:
                    try:
                        UUID(run_name[-36:])
                        run_name = run_name[:-37]  # remove previous uuid
                    except:
                        pass
                run_name += '-' + str(uuid4())
        except Exception:
            pass
        if len(run_name) > 80:
            run_name = run_name[:79]
        return run_name

    def run_workflow(self, input_json, sfn=None,
                     env=None, jobid=None, sleep=3, verbose=True,
                     open_browser=True, dryrun=False):
        '''
        input_json is either a dict or a file
        accession is unique name that we be part of run id
        '''
        if isinstance(input_json, dict):
            data = copy.deepcopy(input_json)
        elif isinstance(input_json, str) and os.path.exists(input_json):
            with open(input_json) as input_file:
                data = json.load(input_file)
        else:
            raise Exception("input json must be either a file or a dictionary")
        # jobid can be specified through run_workflow option (priority),
        # or in input_json
        if not jobid:
            if 'jobid' in data:
                jobid = data['jobid']
            else:
                jobid = create_jobid()
        data['jobid'] = jobid

        if not sfn:
            sfn = self.default_stepfunction_name
        if not env:
            env = self.default_env
        client = boto3.client('stepfunctions', region_name=AWS_REGION)
        base_url = 'https://console.aws.amazon.com/states/home?region=' + AWS_REGION + '#/executions/details/'
        # build from appropriate input json
        # assume run_type and and run_id
        if 'run_name' in data['config']:
            if _tibanna not in data:
                data[_tibanna] = dict()
            data[_tibanna]['run_name'] = data['config']['run_name']
        data = _tibanna_settings(data, force_inplace=True, env=env)
        run_name = self.randomize_run_name(data[_tibanna]['run_name'], sfn)
        data[_tibanna]['run_name'] = run_name
        data['config']['run_name'] = run_name
        # updated arn
        arn = EXECUTION_ARN(run_name, sfn)
        data[_tibanna]['exec_arn'] = arn
        # calculate what the url will be
        url = "%s%s" % (base_url, arn)
        data[_tibanna]['url'] = url
        if 'args' in data:  # unicorn-only
            unicorn_input = UnicornInput(data)
            args = unicorn_input.args
            if args.language.startswith('cwl') and args.cwl_directory_local or \
               args.language in ['wdl', 'wdl_v1', 'wdl_draft2'] and args.wdl_directory_local or \
               args.language == 'snakemake' and args.snakemake_directory_local:
                upload_workflow_to_s3(unicorn_input)
                data['args'] = args.as_dict()  # update args
        # submit job as an execution
        aws_input = json.dumps(data)
        if verbose:
            logger.info("about to start run %s" % run_name)
        # trigger the step function to run
        response = None
        if not dryrun:
            try:
                response = client.start_execution(
                    stateMachineArn=STEP_FUNCTION_ARN(sfn),
                    name=run_name,
                    input=aws_input,
                )
                time.sleep(sleep)
            except Exception as e:
                raise(e)

            # trigger the cost updater step function to run
            try:
                costupdater_input = {
                    "sfn_arn": data[_tibanna]['exec_arn'],
                    "log_bucket": data['config']['log_bucket'],
                    "job_id": data['jobid'],
                    "aws_region": AWS_REGION
                }
                costupdater_input = json.dumps(costupdater_input)
                costupdater_response = client.start_execution(
                    stateMachineArn=STEP_FUNCTION_ARN(sfn + "_costupdater"),
                    name=run_name,
                    input=costupdater_input,
                )
                time.sleep(sleep)
            except Exception as e:
                if 'StateMachineDoesNotExist' in str(e):
                    pass # Tibanna was probably deployed without the cost updater
                else:
                    raise e

        # adding execution info to dynamoDB for fast search by awsem job id
        Job.add_to_dd(jobid, run_name, sfn, data['config']['log_bucket'], verbose=verbose)
        data[_tibanna]['response'] = response
        if verbose:
            # print some info
            logger.info("response from aws was: \n %s" % response)
            logger.info("url to view status:")
            logger.info(data[_tibanna]['url'])
            logger.info("JOBID %s submitted" % data['jobid'])
            logger.info("EXECUTION ARN = %s" % data[_tibanna]['exec_arn'])
            if 'cloudwatch_dashboard' in data['config'] and data['config']['cloudwatch_dashboard']:
                cw_db_url = 'https://console.aws.amazon.com/cloudwatch/' + \
                    'home?region=%s#dashboards:name=awsem-%s' % (AWS_REGION, jobid)
                logger.info("Cloudwatch Dashboard = %s" % cw_db_url)
            if open_browser and shutil.which('open') is not None and not dryrun:
                subprocess.call(["open", data[_tibanna]['url']])
        return data

    def run_batch_workflows(self, input_json_list, sfn=None,
                     env=None, sleep=3, verbose=True, open_browser=True, dryrun=False):
        """given a list of input json, run multiple workflows"""
        run_infos = []
        for input_json in input_json_list:
            run_info = self.run_workflow(input_json, env=env, sfn=sfn, sleep=sleep, verbose=verbose,
                       open_browser=False, dryrun=dryrun)
            run_infos.append(run_info)
        return run_infos

    def check_status(self, exec_arn=None, job_id=None):
        """checking status of an execution.
        """
        return Job(exec_arn=exec_arn, job_id=job_id).check_status()

    def check_costupdater_status(self, job_id=None):
        '''checking status of the cost updater'''
        return Job(job_id=job_id).check_costupdater_status()

    def check_output(self, exec_arn=None, job_id=None):
        """checking status of an execution first and if it's success, get output.
        It works only if the execution info is still in the step function.
        """
        return Job(exec_arn=exec_arn, job_id=job_id).check_output()

    def info(self, job_id):
        '''returns content from dynamodb for a given job id in a dictionary form'''
        return Job.info(job_id)

    def kill(self, exec_arn=None, job_id=None, sfn=None, soft=False):
        """kills a running job, identified by either execution arn (exec_arn) or
        job_id, (or job_id and sfn, for old tibanna versions).
        This function kills both ec2 instance and the step function.
        if soft is set True, it will not directly kill the step function
        but instead sends a abort signal to S3 and let the step fucntion handle
        the signal. The abort signal handling is available for >=1.3.0.
        """
        job = Job(exec_arn=exec_arn, job_id=job_id, sfn=sfn)
        if job.check_status() == 'RUNNING':
            # kill awsem ec2 instance
            ec2 = boto3.client('ec2')
            res = ec2.describe_instances(Filters=[{'Name': 'tag:Name', 'Values': ['awsem-' + job.job_id]}])
            if not res['Reservations']:
                raise Exception("instance not available - if you just submitted the job, try again later")
            instance_id = res['Reservations'][0]['Instances'][0]['InstanceId']
            logger.info("terminating EC2 instance")
            resp_term = ec2.terminate_instances(InstanceIds=[instance_id])
            logger.info("Successfully terminated instance: " + str(resp_term))

            if soft:
                logger.info("sending abort signal to s3")
                put_object_s3('', job.job_id + '.aborted', job.log_bucket,
                              encrypt_s3_upload=True if S3_ENCRYT_KEY_ID else False,
                              kms_key_id=S3_ENCRYT_KEY_ID)
                logger.info("Successfully sent abort signal")
            else:
                # kill step function execution
                logger.info("terminating step function execution")
                sf = boto3.client('stepfunctions')
                resp_sf = sf.stop_execution(executionArn=job.exec_arn, error="Aborted")
                logger.info("Successfully terminated step function execution: " + str(resp_sf))

    def kill_all(self, sfn=None):
        """killing all the running jobs"""
        if not sfn:
            sfn = self.default_stepfunction_name
        client = boto3.client('stepfunctions')
        stateMachineArn = STEP_FUNCTION_ARN(sfn)
        res = client.list_executions(stateMachineArn=stateMachineArn, statusFilter='RUNNING')
        while True:
            if 'executions' not in res or not res['executions']:
                break
            for exc in res['executions']:
                self.kill(exc['executionArn'])
            if 'nextToken' in res:
                res = client.list_executions(nextToken=res['nextToken'],
                                             stateMachineArn=stateMachineArn, statusFilter='RUNNING')
            else:
                break

    def log(self, exec_arn=None, job_id=None, exec_name=None, sfn=None,
            postrunjson=False, runjson=False, top=False, top_latest=False,
            inputjson=False, logbucket=None, quiet=False):
        if postrunjson:
            suffix = '.postrun.json'
        elif runjson:
            suffix = '.run.json'
        elif top:
            suffix = '.top'
        elif top_latest:
            suffix = '.top_latest'
        elif inputjson:
            suffix = '.input.json'
        else:
            suffix = '.log'
        if not sfn:
            sfn = self.default_stepfunction_name
        sf = boto3.client('stepfunctions')
        if not exec_arn and exec_name:
            exec_arn = EXECUTION_ARN(exec_name, sfn)
        job = Job(exec_arn=exec_arn, job_id=job_id, sfn=sfn)
        try:
            res_s3 = boto3.client('s3').get_object(Bucket=job.log_bucket, Key=job.job_id + suffix)
        except Exception as e:
            if 'NoSuchKey' in str(e):
                if not quiet:
                    logger.info("log/postrunjson file is not ready yet. " +
                                "Wait a few seconds/minutes and try again.")
                return ''
            else:
                raise e
        if res_s3:
            return(res_s3['Body'].read().decode('utf-8', 'backslashreplace'))
        return None

    def stat(self, sfn=None, status=None, verbose=False, n=None, job_ids=None):
        """print out executions with details (-v)
        status can be one of 'RUNNING'|'SUCCEEDED'|'FAILED'|'TIMED_OUT'|'ABORTED'
        or specify a list of job ids
        """
        if n and job_ids:
            raise Exception("n and job_id filters do not work together.")
        if sfn and job_ids:
            raise Exception("Please do not specify sfn when job_ids are specified.")
        if status and job_ids:
            raise Exception("Status filter cannot be specified when job_ids are specified.")
        if verbose:
            print("{}\t{}\t{}\t{}\t{}\t{}\t{}\t{}\t{}\t{}\t{}".format('jobid', 'status', 'name',
                                                                      'start_time', 'stop_time',
                                                                      'instance_id', 'instance_type',
                                                                      'instance_status', 'ip', 'key',
                                                                      'password'))
        else:
            print("{}\t{}\t{}\t{}\t{}".format('jobid', 'status', 'name', 'start_time', 'stop_time'))
        client = boto3.client('stepfunctions')
        ec2 = boto3.client('ec2')

        def parse_exec_desc_and_ec2_desc(exec_arn, verbose):
            # collecting execution stats
            exec_desc = client.describe_execution(executionArn=exec_arn)

            # getting info from execution description
            exec_desc = client.describe_execution(executionArn=exec_arn)
            job_id = json.loads(exec_desc['input']).get('jobid', 'no jobid')
            status = exec_desc['status']
            name = exec_desc['name']
            start_time = exec_desc['startDate'].strftime("%Y-%m-%d %H:%M")
            if 'stopDate' in exec_desc:
                stop_time = exec_desc['stopDate'].strftime("%Y-%m-%d %H:%M")
            else:
                stop_time = ''

            # collect instance stats
            ec2_desc = ec2.describe_instances(Filters=[{'Name': 'tag:Name', 'Values': ['awsem-' + job_id]}])

            # getting info from ec2 description
            if ec2_desc['Reservations']:
                ec2_desc_inst = ec2_desc['Reservations'][0]['Instances'][0]
                instance_status = ec2_desc_inst['State']['Name']
                instance_id = ec2_desc_inst['InstanceId']
                instance_type = ec2_desc_inst['InstanceType']
                if instance_status not in ['terminated', 'shutting-down']:
                    instance_ip = ec2_desc_inst.get('PublicIpAddress', '-')
                    keyname = ec2_desc_inst.get('KeyName', '-')
                    password = json.loads(exec_desc['input'])['config'].get('password', '-')
                else:
                    instance_ip = '-'
                    keyname = '-'
                    password = '-'
            else:
                instance_status = '-'
                instance_id = '-'
                instance_type = '-'
                instance_ip = '-'
                keyname = '-'
                password = '-'

            parsed_stat = (job_id, status, name, start_time, stop_time,
                           instance_id, instance_type, instance_status,
                           instance_ip, keyname, password)
            if verbose:
                print("{}\t{}\t{}\t{}\t{}\t{}\t{}\t{}\t{}\t{}\t{}".format(*parsed_stat))
            else:
                print("{}\t{}\t{}\t{}\t{}".format(*parsed_stat[0:5]))

        if job_ids:
            for job_id in job_ids:
                dd_info = self.info(job_id)
                if 'Execution Name' not in dd_info:
                    raise Exception("Cannot find execution name for job ID %s" % job_id)
                if 'Step Function' not in dd_info:
                    raise Exception("Cannot find step function for job ID %s" % job_id)
                exec_arn = EXECUTION_ARN(dd_info['Execution Name'], dd_info['Step Function'])
                parse_exec_desc_and_ec2_desc(exec_arn, verbose)
        else:
            if not sfn:
                sfn = self.default_stepfunction_name
            args = {
                'stateMachineArn': STEP_FUNCTION_ARN(sfn),
                'maxResults': 100
            }
            if status:
                args['statusFilter'] = status
            res = dict()
            res = client.list_executions(**args)
            k = 0
            while True:
                if n and k == n:
                    break
                if 'executions' not in res or not res['executions']:
                    break
                for exc in res['executions']:
                    if n and k == n:
                        break
                    k = k + 1
                    parse_exec_desc_and_ec2_desc(exc['executionArn'], verbose)
                if 'nextToken' in res:
                    res = client.list_executions(nextToken=res['nextToken'], **args)
                else:
                    break

    def list_sfns(self, numbers=False):
        """list all step functions, optionally with a summary (-n)"""
        st = boto3.client('stepfunctions')
        res = st.list_state_machines(
            maxResults=1000
        )
        header = "name\tcreation_date"
        if numbers:
            header = header + "\trunning\tsucceeded\tfailed\taborted\ttimed_out"
        print(header)
        for s in res['stateMachines']:
            if not s['name'].startswith('tibanna_' + self.sfn_type):
                continue
            line = "%s\t%s" % (s['name'], str(s['creationDate']))
            if numbers:
                counts = self.count_status(s['stateMachineArn'], st)
                for status in ['RUNNING', 'SUCCEEDED', 'FAILED', 'ABORTED', 'TIMED_OUT']:
                    line = line + "\t%i" % counts[status]
            print(line)

    def count_status(self, sfn_arn, client):
            next_token = None
            count = dict()
            while True:
                args = {'stateMachineArn': sfn_arn,
                        # 'statusFilter': status,
                        'maxResults': 1000}
                if next_token:
                    args['nextToken'] = next_token
                res = client.list_executions(**args)
                for status in ['RUNNING', 'SUCCEEDED', 'FAILED', 'ABORTED', 'TIMED_OUT']:
                    count[status] = count.get(status, 0) + sum([r['status'] == status for r in res['executions']])
                if res.get('nextToken', ''):
                    next_token = res['nextToken']
                else:
                    break
            return count

    def clear_input_json_template(self, input_json_template):
        """clear awsem template for reuse"""
        if 'jobid' in input_json_template:
            del input_json_template['jobid']
        if 'response' in input_json_template['_tibanna']:
            del(input_json_template['_tibanna']['response'])
        if 'run_name' in input_json_template['_tibanna'] and len(input_json_template['_tibanna']['run_name']) > 40:
            input_json_template['_tibanna']['run_name'] = input_json_template['_tibanna']['run_name'][:-36]
            input_json_template['config']['run_name'] = input_json_template['_tibanna']['run_name']

    def rerun(self, exec_arn=None, job_id=None, sfn=None,
              override_config=None, app_name_filter=None,
              instance_type=None, shutdown_min=None, ebs_size=None, ebs_type=None, ebs_iops=None, ebs_throughput=None,
              overwrite_input_extra=None, key_name=None, name=None, use_spot=None, do_not_use_spot=None):
        """rerun a specific job
        override_config : dictionary for overriding config (keys are the keys inside config)
            e.g. override_config = { 'instance_type': 't2.micro' }
        app_name_filter : app_name (e.g. hi-c-processing-pairs), if specified,
        then rerun only if it matches app_name
        """
        if not sfn:
            sfn = self.default_stepfunction_name  # this is a target sfn
        if not exec_arn and job_id:
            input_json_template = json.loads(self.log(job_id=job_id, inputjson=True))
        else:
            client = boto3.client('stepfunctions')
            res = client.describe_execution(executionArn=exec_arn)
            input_json_template = json.loads(res['input'])
        # filter by app_name
        if app_name_filter:
            if 'app_name' not in input_json_template:
                return(None)
            if input_json_template['app_name'] != app_name_filter:
                return(None)
        self.clear_input_json_template(input_json_template)
        # override config
        if not override_config:
            override_config = dict()
            if instance_type:
                override_config['instance_type'] = instance_type
            if shutdown_min:
                override_config['shutdown_min'] = shutdown_min
            if ebs_size:
                if ebs_size.endswith('x'):
                    override_config['ebs_size'] = ebs_size
                else:
                    override_config['ebs_size'] = int(ebs_size)
            if overwrite_input_extra:
                override_config['overwrite_input_extra'] = overwrite_input_extra
            if key_name:
                override_config['key_name'] = key_name
            if ebs_type:
                override_config['ebs_type'] = ebs_type
                if ebs_type == 'gp2':
                    override_config['ebs_iops'] = ''
            if ebs_iops:
                override_config['ebs_iops'] = ebs_iops
            if ebs_throughput:
                override_config['ebs_throughput'] = ebs_throughput
            if use_spot:
                override_config['spot_instance'] = True
            if do_not_use_spot:
                override_config['spot_instance'] = False
        if override_config:
            for k, v in iter(override_config.items()):
                input_json_template['config'][k] = v
        return(self.run_workflow(input_json_template, sfn=sfn))

    def rerun_many(self, sfn=None, stopdate='13Feb2018', stophour=13,
                   stopminute=0, offset=0, sleeptime=5, status='FAILED',
                   override_config=None, app_name_filter=None,
                   instance_type=None, shutdown_min=None, ebs_size=None, ebs_type=None, ebs_iops=None, ebs_throughput=None,
                   overwrite_input_extra=None, key_name=None, name=None, use_spot=None, do_not_use_spot=None):
        """Reruns step function jobs that failed after a given time point
        (stopdate, stophour (24-hour format), stopminute)
        By default, stophour should be the same as your system time zone.
        This can be changed by setting a different offset.
        If offset=5, for instance, that means your stoptime=12 would correspond
        to your system time=17. Sleeptime is sleep time in seconds between rerun submissions.
        By default, it reruns only 'FAILED' runs, but this can be changed by resetting status.
        examples)
        rerun_many('tibanna_pony-dev')
        rerun_many('tibanna_pony', stopdate= '14Feb2018', stophour=14, stopminute=20)
        """
        if not sfn:
            sfn = self.default_stepfunction_name
        stophour = stophour + offset
        stoptime = stopdate + ' ' + str(stophour) + ':' + str(stopminute)
        stoptime_in_datetime = datetime.strptime(stoptime, '%d%b%Y %H:%M')
        client = boto3.client('stepfunctions')
        sflist = client.list_executions(stateMachineArn=STEP_FUNCTION_ARN(sfn), statusFilter=status)
        k = 0
        for exc in sflist['executions']:
            if exc['stopDate'].replace(tzinfo=None) > stoptime_in_datetime:
                k = k + 1
                self.rerun(exc['executionArn'], sfn=sfn,
                           override_config=override_config, app_name_filter=app_name_filter,
                           instance_type=instance_type, shutdown_min=shutdown_min, ebs_size=ebs_size,
                           ebs_type=ebs_type, ebs_iops=ebs_iops, ebs_throughput=ebs_throughput,
                           overwrite_input_extra=overwrite_input_extra, key_name=key_name, name=name,
                           use_spot=use_spot, do_not_use_spot=do_not_use_spot)
                time.sleep(sleeptime)

    def env_list(self, name):
        # don't set this as a global, since not all tasks require it
        envlist = {
            self.run_task_lambda: {'AMI_ID': self.ami_id,
                                   'TIBANNA_REPO_NAME': TIBANNA_REPO_NAME,
                                   'TIBANNA_REPO_BRANCH': TIBANNA_REPO_BRANCH},
            self.check_task_lambda: {'TIBANNA_DEFAULT_STEP_FUNCTION_NAME': self.default_stepfunction_name}
        }
        if TIBANNA_PROFILE_ACCESS_KEY and TIBANNA_PROFILE_SECRET_KEY:
            envlist[self.run_task_lambda].update({
                'TIBANNA_PROFILE_ACCESS_KEY': TIBANNA_PROFILE_ACCESS_KEY,
                'TIBANNA_PROFILE_SECRET_KEY': TIBANNA_PROFILE_SECRET_KEY}
            )
        return envlist.get(name, '')

    @staticmethod
    def add_role_to_kms(*, kms_key_id, role_arn):
        """ Adds role_arn to the given kms_key_id policy.
            Note that tibanna assumes a KMS key and existing policy.
            This function will have no effect if no policy exists with
            the sid: 'Allow use of the key'
        """
        # adjust KMS key policy to allow this role to use s3 key
        kms_client = boto3.client('kms')
        policy = kms_client.get_key_policy(KeyId=kms_key_id, PolicyName='default')
        policy = json.loads(policy['Policy'])
        for statement in policy['Statement']:
            if statement.get('Sid', '') == 'Allow use of the key':  # 4dn-cloud-infra dependency -Will Dec 1 2021
                if isinstance(statement['Principal']['AWS'], str):
                    statement['Principal']['AWS'] = [statement['Principal']['AWS']]
                statement['Principal']['AWS'].append(role_arn)
                break
        policy = json.dumps(policy)
        kms_client.put_key_policy(
            KeyId=kms_key_id,
            PolicyName='default',
            Policy=policy
        )

    @staticmethod
    def cleanup_kms():
        """ Cleans up the KMS access policy by removing revoked IAM roles.
            These revoked entities show up in KMS with prefix AROA. This function
            removes those garbage values.
        """
        if not S3_ENCRYT_KEY_ID:
            return

        kms_client = boto3.client('kms')
        policy = kms_client.get_key_policy(KeyId=S3_ENCRYT_KEY_ID, PolicyName='default')
        policy = json.loads(policy['Policy'])
        for statement in policy['Statement']:
            if statement.get('Sid', '') == 'Allow use of the key' or statement.get('Sid', '') == 'Allow attachment of persistent resources':  # default
                iam_entities = statement['Principal'].get('AWS', [])
                if type(iam_entities) == str:
                    iam_entities = [iam_entities]
                filtered_entities = list(filter(lambda s: not s.startswith('AROA'), iam_entities))
                statement['Principal']['AWS'] = filtered_entities
                break
        policy = json.dumps(policy)
        kms_client.put_key_policy(
            KeyId=S3_ENCRYT_KEY_ID,
            PolicyName='default',
            Policy=policy
        )

    def deploy_lambda(self, name, suffix, usergroup='', quiet=False, subnets=None, security_groups=None):
        """
        deploy a single lambda using the aws_lambda.deploy_function (BETA).
        subnets and security groups are lists of subnet IDs and security group IDs, in case
        you want the lambda to be deployed into specific subnets / security groups.
        """
        import aws_lambda
        if name not in dir(self.lambdas_module):
            raise Exception("Could not find lambda function file for %s" % name)
        lambda_fxn_module = importlib.import_module('.'.join([self.lambdas_module.__name__,  name]))
        requirements_fpath = os.path.join(self.lambdas_module.__path__[0], 'requirements.txt')
        # add extra config to the lambda deployment
        extra_config = {}
        envs = self.env_list(name)
        if envs:
            extra_config['Environment'] = {'Variables': envs}
        if subnets or security_groups:
            extra_config['VpcConfig'] = {}
            if not 'Environment' in extra_config:
                extra_config['Environment'] = {'Variables': {}}
            if subnets:
                extra_config['VpcConfig'].update({'SubnetIds': subnets})
                extra_config['Environment']['Variables'].update({'SUBNETS': ','.join(subnets)})
            if security_groups:
                extra_config['VpcConfig'].update({'SecurityGroupIds': security_groups})
                extra_config['Environment']['Variables'].update({'SECURITY_GROUPS': ','.join(security_groups)})
        if S3_ENCRYT_KEY_ID:
            extra_config['Environment']['Variables'].update({'S3_ENCRYPT_KEY_ID': S3_ENCRYT_KEY_ID})
        tibanna_iam = self.IAM(usergroup)
        if name == self.run_task_lambda:
            extra_config['Environment']['Variables']['AWS_S3_ROLE_NAME'] \
                = tibanna_iam.role_name('ec2')

        # Add Tibanna version to the env variables on deploy, so that they don't have to parse the pyproject.toml
        extra_config['Environment']['Variables']['TIBANNA_VERSION'] = __version__

        # add role
        logger.info('name=%s' % name)
        role_arn_prefix = 'arn:aws:iam::' + AWS_ACCOUNT_NUMBER + ':role/'
        role_arn = role_arn_prefix + tibanna_iam.role_name(name)
        logger.info("role_arn=" + role_arn)
        extra_config['Role'] = role_arn

        # first delete the existing function to avoid the weird AWS KMS lambda error
        function_name_prefix = getattr(self.lambdas_module, name).config.get('function_name')
        function_name_suffix = create_tibanna_suffix(suffix, usergroup)
        full_function_name = function_name_prefix + function_name_suffix
        if name not in self.do_not_delete:
            try:
                if quiet:
                    res = boto3.client('lambda').get_function(FunctionName=full_function_name)
                    logger.info("deleting existing lambda")
                    res = boto3.client('lambda').delete_function(FunctionName=full_function_name)
                else:
                    boto3.client('lambda').get_function(FunctionName=full_function_name)
                    logger.info("deleting existing lambda")
                    boto3.client('lambda').delete_function(FunctionName=full_function_name)
            except Exception as e:
                if 'Function not found' in str(e):
                    pass

        if quiet:
            res = aws_lambda.deploy_function(lambda_fxn_module,
                                             function_name_suffix=function_name_suffix,
                                             package_objects=self.tibanna_packages,
                                             requirements_fpath=requirements_fpath,
                                             extra_config=extra_config)
        else:
            aws_lambda.deploy_function(lambda_fxn_module,
                                       function_name_suffix=function_name_suffix,
                                       package_objects=self.tibanna_packages,
                                       requirements_fpath=requirements_fpath,
                                       extra_config=extra_config)

    def deploy_core(self, name, suffix=None, usergroup='', subnets=None, security_groups=None,
                    quiet=False):
        """deploy/update lambdas only"""
        logger.info("preparing for deploy...")
        if name == 'all':
            names = self.lambda_names
        elif name == 'unicorn':
            names = self.UNICORN_LAMBDAS
        else:
            names = [name, ]
        for name in names:
            self.deploy_lambda(name, suffix, usergroup, subnets=subnets, security_groups=security_groups,
                               quiet=quiet)

    def setup_tibanna_env(self, buckets='', usergroup_tag='default', no_randomize=False,
                          do_not_delete_public_access_block=False, verbose=False):
        """set up usergroup environment on AWS
        This function is called automatically by deploy_tibanna or deploy_unicorn
        Use it only when the IAM permissions need to be reset"""
        logger.info("setting up tibanna usergroup environment on AWS...")
        if not AWS_ACCOUNT_NUMBER or not AWS_REGION:
            logger.info("Please set and export environment variable AWS_ACCOUNT_NUMBER and AWS_REGION!")
            exit(1)
        if not buckets:
            logger.warning("Without setting buckets (using --buckets)," +
                           "Tibanna would have access to only public buckets." +
                           "To give permission to Tibanna for private buckets," +
                           "use --buckets=<bucket1>,<bucket2>,...")
            time.sleep(2)
        if buckets:
            bucket_names = buckets.split(',')
        else:
            bucket_names = None
        if bucket_names and not do_not_delete_public_access_block:
            client = boto3.client('s3')
            for b in bucket_names:
                logger.info("Deleting public access block for bucket %s" % b)
                try:
                    response = client.delete_public_access_block(Bucket=b)
                except Exception as e:
                    logger.warning("public access block could not be deleted on bucket %s - skipping.." % b)
        tibanna_iam = self.IAM(usergroup_tag, bucket_names, no_randomize=no_randomize)
        tibanna_iam.create_tibanna_iam(verbose=verbose)
        logger.info("Tibanna usergroup %s has been created on AWS." % tibanna_iam.user_group_name)
        return tibanna_iam.user_group_name

    def deploy_tibanna(self, suffix=None, usergroup='', setup=False, no_randomize=False,
                       default_usergroup_tag='default',
                       buckets='', setenv=False, do_not_delete_public_access_block=False,
                       deploy_costupdater=False, subnets=None, security_groups=None, quiet=False):
        """deploy tibanna unicorn or pony to AWS cloud (pony is for 4DN-DCIC only)"""
        if setup:
            if usergroup:
                usergroup = self.setup_tibanna_env(buckets, usergroup, True,
                            do_not_delete_public_access_block=do_not_delete_public_access_block)
            else:  # override usergroup
                usergroup = self.setup_tibanna_env(buckets, usergroup_tag=default_usergroup_tag,
                            do_not_delete_public_access_block=do_not_delete_public_access_block,
                            no_randomize=no_randomize)

        # If deploying with encryption, revoke KMS perms from previous deployment.
        # Note that this means if you deploy while tibanna is running on an encrypted env,
        # it will stop working! - Will Dec 2 2021
        if S3_ENCRYT_KEY_ID:
            self.cleanup_kms()

        # this function will remove existing step function on a conflict
        step_function_name = self.create_stepfunction(suffix, usergroup=usergroup)
        logger.info("creating a new step function... %s" % step_function_name)

        if(deploy_costupdater):
            step_function_cu_name = self.create_stepfunction(suffix, usergroup=usergroup, costupdater=True)
            logger.info("creating a new step function... %s" % step_function_cu_name)

        if setenv:
            os.environ['TIBANNA_DEFAULT_STEP_FUNCTION_NAME'] = step_function_name
            with open(os.getenv('HOME') + "/.bashrc", "a") as outfile:  # 'a' stands for "append"
                outfile.write("\nexport TIBANNA_DEFAULT_STEP_FUNCTION_NAME=%s\n" % step_function_name)

        logger.info("deploying lambdas...")
        self.deploy_core('all', suffix=suffix, usergroup=usergroup, subnets=subnets,
                         security_groups=security_groups, quiet=quiet)

        if(deploy_costupdater):
            self.deploy_lambda(self.update_cost_lambda, suffix=suffix, usergroup=usergroup,
                               subnets=subnets, security_groups=security_groups, quiet=quiet)

        # Give new roles KMS permissions
        if S3_ENCRYT_KEY_ID:
            self.cleanup_kms()  # cleanup again just in case
            tibanna_iam = self.IAM(usergroup)
            for role_type in tibanna_iam.role_types:
                role_name = tibanna_iam.role_name(role_type)
                self.add_role_to_kms(kms_key_id=S3_ENCRYT_KEY_ID,
                                     role_arn=f'arn:aws:iam::{tibanna_iam.account_id}:role/{role_name}')

        dd_utils.create_dynamo_table(DYNAMODB_TABLE, DYNAMODB_KEYNAME)
        return step_function_name

    def deploy_unicorn(self, suffix=None, no_setup=False, buckets='',
                       no_setenv=False, usergroup='', do_not_delete_public_access_block=False,
                       deploy_costupdater=False, subnets=None, security_groups=None,
                       quiet=False):
        """deploy tibanna unicorn to AWS cloud"""
        self.deploy_tibanna(suffix=suffix, usergroup=usergroup, setup=not no_setup,
                            buckets=buckets, setenv=not no_setenv,
                            do_not_delete_public_access_block=do_not_delete_public_access_block,
                            deploy_costupdater=deploy_costupdater, subnets=subnets,
                            security_groups=security_groups, quiet=quiet)

    def add_user(self, user, usergroup):
        """add a user to a tibanna group"""
        groupname_prefix = 'tibanna_'
        if self.lambda_type:
            groupname_prefix += self.lambda_type + '_'
        boto3.client('iam').add_user_to_group(
            GroupName=groupname_prefix + usergroup,
            UserName=user
        )

    def users(self):
        """list all users along with their associated tibanna user groups"""
        client = boto3.client('iam')
        marker = None
        while True:
            if marker:
                res = client.list_users(Marker=marker)
            else:
                res = client.list_users()
            print("user\ttibanna_usergroup")
            for r in res['Users']:
                res_groups = client.list_groups_for_user(
                    UserName=r['UserName']
                )
                groups = [rg['GroupName'] for rg in res_groups['Groups']]
                groups = filter(lambda x: 'tibanna_' in x, groups)
                groups = [x.replace('tibanna_', '') for x in groups]
                print("%s\t%s" % (r['UserName'], ','.join(groups)))
            marker = res.get('Marker', '')
            if not marker:
                break

    def create_stepfunction(self, dev_suffix=None,
                            region_name=AWS_REGION,
                            aws_acc=AWS_ACCOUNT_NUMBER,
                            usergroup=None,
                            costupdater=False):
        if not aws_acc or not region_name:
            logger.info("Please set and export environment variable AWS_ACCOUNT_NUMBER and AWS_REGION!")
            exit(1)
        # create a step function definition object
        if(costupdater):
            sfndef = self.StepFunctionCU(dev_suffix, region_name, aws_acc, usergroup)
        else:
            sfndef = self.StepFunction(dev_suffix, region_name, aws_acc, usergroup)
        # if this encouters an existing step function with the same name, delete
        sfn = boto3.client('stepfunctions', region_name=region_name)
        retries = 12  # wait 10 seconds between retries for total of 120s
        for i in range(retries):
            try:
                sfn.create_state_machine(
                    name=sfndef.sfn_name,
                    definition=json.dumps(sfndef.definition, indent=4, sort_keys=True),
                    roleArn=sfndef.sfn_role_arn
                )
            except sfn.exceptions.StateMachineAlreadyExists as e:
                # get ARN from the error and format as necessary
                exc_str = str(e)
                if 'State Machine Already Exists:' not in exc_str:
                    logger.error('Cannot delete state machine. Exiting...' % exc_str)
                    raise(e)
                sfn_arn = exc_str.split('State Machine Already Exists:')[-1].strip().strip("''")
                logger.info('Step function with name %s already exists!' % sfndef.sfn_name)
                logger.info('Updating the state machine...')
                try:
                    sfn.update_state_machine(
                        stateMachineArn=sfn_arn,
                        definition=json.dumps(sfndef.definition, indent=4, sort_keys=True),
                        roleArn=sfndef.sfn_role_arn
                    )
                except Exception as e:
                    logger.error('Error updating state machine %s' % str(e))
                    raise(e)
            except Exception as e:
                raise(e)
            break
        return sfndef.sfn_name

    def check_metrics_plot(self, job_id, log_bucket):
        return True if does_key_exist(log_bucket, job_id + '.metrics/metrics.html', quiet=True) else False

    def check_metrics_lock(self, job_id, log_bucket):
        return True if does_key_exist(log_bucket, job_id + '.metrics/lock', quiet=True) else False

    def plot_metrics(self, job_id, sfn=None, directory='.', open_browser=True, force_upload=False,
                     update_html_only=False, endtime='', filesystem='/dev/nvme1n1', instance_id=''):
        ''' retrieve instance_id and plots metrics '''
        if not sfn:
            sfn = self.default_stepfunction_name
        postrunjsonstr = self.log(job_id=job_id, sfn=sfn, postrunjson=True, quiet=True)
        if postrunjsonstr:
            postrunjson = AwsemPostRunJson(**json.loads(postrunjsonstr))
            job = postrunjson.Job
            if hasattr(job, 'end_time_as_datetime') and job.end_time_as_datetime:
                job_complete = True
            else:
                job_complete = False
            log_bucket = postrunjson.config.log_bucket
            instance_type = postrunjson.config.instance_type or 'unknown'
        else:
            runjsonstr = self.log(job_id=job_id, sfn=sfn, runjson=True, quiet=True)
            job_complete = False
            if runjsonstr:
                runjson = AwsemRunJson(**json.loads(runjsonstr))
                job = runjson.Job
                log_bucket = runjson.config.log_bucket
                instance_type = runjson.config.instance_type or 'unknown'
            else:
                raise Exception("Neither postrun json nor run json can be retrieved." +
                                "Check job_id or step function?")
        # report already on s3 with a lock
        if self.check_metrics_plot(job_id, log_bucket) and \
           self.check_metrics_lock(job_id, log_bucket) and \
           not force_upload:
            logger.info("Metrics plot is already on S3 bucket.")
            logger.info('metrics url= ' + METRICS_URL(log_bucket, job_id))
            # open metrics html in browser
            if open_browser:
                webbrowser.open(METRICS_URL(log_bucket, job_id))
            return None
        # report not already on s3 with a lock
        starttime = job.start_time_as_datetime
        if not endtime:
            if hasattr(job, 'end_time_as_datetime') and job.end_time_as_datetime:
                endtime = job.end_time_as_datetime
            else:
                endtime = datetime.utcnow()
        if hasattr(job, 'filesystem') and job.filesystem:
            filesystem = job.filesystem
        else:
            filesystem = filesystem
        if not instance_id:
            if hasattr(job, 'instance_id') and job.instance_id:
                instance_id = job.instance_id
            else:
                ddres = dict()
                try:
                    dd = boto3.client('dynamodb')
                    ddres = dd.query(TableName=DYNAMODB_TABLE,
                                     KeyConditions={'Job Id': {'AttributeValueList': [{'S': job_id}],
                                                               'ComparisonOperator': 'EQ'}})
                except Exception as e:
                    pass
                if 'Items' in ddres:
                    instance_id = ddres['Items'][0].get('instance_id', {}).get('S', '')
                if not instance_id:
                    ec2 = boto3.client('ec2')
                    res = ec2.describe_instances(Filters=[{'Name': 'tag:Name', 'Values': ['awsem-' + job_id]}])
                    if res['Reservations']:
                        instance_id = res['Reservations'][0]['Instances'][0]['InstanceId']
                        instance_status = res['Reservations'][0]['Instances'][0]['State']['Name']
                        if instance_status in ['terminated', 'shutting-down']:
                            job_complete = True  # job failed
                        else:
                            job_complete = False  # still running
                    else:
                        # waiting 10 min to be sure the istance is starting
                        if (datetime.utcnow() - starttime) / timedelta(minutes=1) < 5:
                            raise Exception("the instance is still setting up. " +
                                            "Wait a few seconds/minutes and try again.")
                        else:
                            raise Exception("instance id not available for this run. Try manually providing " + \
                                            "it using the instance_id parameter (--instance-id option)")
        # plotting
        if update_html_only:
            self.TibannaResource.update_html(log_bucket, job_id + '.metrics/')
        else:
            try:
                # We don't want to recompute the cost estimate - we take it from the tsv
                cost_estimate, cost_estimate_type = get_cost_estimate_from_tsv(log_bucket, job_id)
                if(cost_estimate == 0.0): # Cost estimate is not yet in tsv -> compute it
                    cost_estimate, cost_estimate_type = self.cost_estimate(job_id=job_id)

                M = self.TibannaResource(instance_id, filesystem, starttime, endtime, cost_estimate = cost_estimate, cost_estimate_type=cost_estimate_type)
                top_content = self.log(job_id=job_id, top=True)
                M.plot_metrics(instance_type, directory, top_content=top_content)
            except Exception as e:
                raise MetricRetrievalException(e)
            # upload files
            M.upload(bucket=log_bucket, prefix=job_id + '.metrics/', lock=job_complete)
            # clean up uploaded files
            for f in M.list_files:
                os.remove(f)
        logger.info('metrics url= ' + METRICS_URL(log_bucket, job_id))
        # open metrics html in browser
        if open_browser:
            webbrowser.open(METRICS_URL(log_bucket, job_id))

    def cost_estimate(self, job_id, update_tsv=False, force=False):
        postrunjsonstr = self.log(job_id=job_id, postrunjson=True)
        if not postrunjsonstr:
            logger.info("Cost estimation error: postrunjson not found")
            return 0.0, "NA"
        postrunjsonobj = json.loads(postrunjsonstr)
        postrunjson = AwsemPostRunJson(**postrunjsonobj)
        log_bucket = postrunjson.config.log_bucket
        encryption = postrunjson.config.encrypt_s3_upload
        kms_key_id = postrunjson.config.kms_key_id

        # We return the real cost, if it is available, but don't automatically update the Cost row in the tsv
        if not force:
            precise_cost = self.cost(job_id, update_tsv=False)
            if(precise_cost and precise_cost > 0.0):
                if update_tsv:
                    update_cost_estimate_in_tsv(log_bucket, job_id, precise_cost, cost_estimate_type="actual cost",
                                                encryption=encryption, kms_key_id=kms_key_id)
                return precise_cost, "actual cost"

        # awsf_image was added in 1.0.0. We use that to get the correct ebs root type
        ebs_root_type = 'gp3' if 'awsf_image' in postrunjsonobj['config'] else 'gp2'

        cost_estimate, cost_estimate_type = get_cost_estimate(postrunjson, ebs_root_type)

        if update_tsv:
            update_cost_estimate_in_tsv(log_bucket, job_id, cost_estimate, cost_estimate_type,
                                        encryption=encryption, kms_key_id=kms_key_id)
        return cost_estimate, cost_estimate_type

    def cost(self, job_id, sfn=None, update_tsv=False):
        if not sfn:
            sfn = self.default_stepfunction_name
        postrunjsonstr = self.log(job_id=job_id, sfn=sfn, postrunjson=True)
        if not postrunjsonstr:
            return None
        postrunjson = AwsemPostRunJson(**json.loads(postrunjsonstr))

        cost = get_cost(postrunjson, job_id)

        if update_tsv:
            log_bucket = postrunjson.config.log_bucket
            encryption = postrunjson.config.encrypt_s3_upload
            kms_key_id = postrunjson.config.kms_key_id
            update_cost_in_tsv(log_bucket, job_id, cost,
                               encryption=encryption, kms_key_id=kms_key_id)

        return cost


    def does_dynamo_table_exist(self, tablename):
        try:
            res = boto3.client('dynamodb').describe_table(
                TableName=tablename
            )
            if res:
                return True
            else:
                raise Exception("error describing table %s" % tablename)
        except Exception as e:
            if 'Requested resource not found' in str(e):
                return False
            else:
                raise Exception("error describing table %s" % tablename)

    def create_dynamo_table(self, tablename, keyname):
        if self.does_dynamo_table_exist(tablename):
            logger.info("dynamodb table %s already exists. skip creating db" % tablename)
        else:
            response = boto3.client('dynamodb').create_table(
                TableName=tablename,
                AttributeDefinitions=[
                    {
                         'AttributeName': keyname,
                         'AttributeType': 'S'
                    }
                ],
                KeySchema=[
                    {
                        'AttributeName': keyname,
                        'KeyType': 'HASH'
                     }
                ],
                BillingMode='PAY_PER_REQUEST'
            )

    def is_idle(self, instance_id, max_cpu_percent_threshold=1.0):
        """returns True if the instance is idle i.e. not doing anything for
        the past 1 hour and is safe to kill"""
        end = datetime.now(tzutc())
        start = end - timedelta(hours=1)
        filesystem = '/dev/nvme1n1'  # doesn't matter for cpu utilization
        try:
            cw_res = self.TibannaResource(instance_id, filesystem, start, end).as_dict()
        except Exception as e:
            raise MetricRetrievalException(e)
        if not cw_res['max_cpu_utilization_percent']:
            return True
        if cw_res['max_cpu_utilization_percent'] < max_cpu_percent_threshold:
            return True
        return False

    def cleanup(self, user_group_name, suffix='', ignore_errors=True, do_not_remove_iam_group=False,
                purge_history=False, verbose=False):

        def handle_error(errmsg):
            if ignore_errors:
                if verbose:
                    logger.warning(errmsg)
                    logger.info("continue to remove the other components")
            else:
                raise Exception(errmsg)

        if user_group_name.startswith('tibanna_'):
            raise Exception("User_group_name does not start with tibanna or tibanna_unicorn.")
        if suffix:
            lambda_suffix = '_' + user_group_name + '_' + suffix
        else:
            lambda_suffix = '_' + user_group_name
        # delete step functions
        sfn = 'tibanna_' + self.sfn_type + lambda_suffix
        if verbose:
            logger.info("deleting step function %s" % sfn)
        try:
            boto3.client('stepfunctions').delete_state_machine(stateMachineArn=STEP_FUNCTION_ARN(sfn))
        except Exception as e:
            handle_error("Failed to cleanup step function: %s" % str(e))

        sfn_costupdater = 'tibanna_' + self.sfn_type + lambda_suffix + '_costupdater'
        if verbose:
            logger.info("deleting step function %s" % sfn_costupdater)
        try:
            # This does not produce an exception, if the step function does not exist
            boto3.client('stepfunctions').delete_state_machine(stateMachineArn=STEP_FUNCTION_ARN(sfn_costupdater))
        except Exception as e:
            handle_error("Failed to cleanup step function: %s" % str(e))

        # delete lambdas
        lambda_client = boto3.client('lambda')
        for lmb in self.lambda_names:
            if verbose:
                logger.info("deleting lambda functions %s" % lmb + lambda_suffix)
            try:
                lambda_client.delete_function(FunctionName=lmb + lambda_suffix)
            except Exception as e:
                handle_error("Failed to cleanup lambda: %s" % str(e))

        if verbose:
            logger.info("deleting lambda functions %s" % self.update_cost_lambda + lambda_suffix)
        try:
            lambda_client.delete_function(FunctionName=self.update_cost_lambda  + lambda_suffix)
        except Exception as e:
            handle_error("Failed to cleanup lambda: %s" % str(e))

        # delete IAM policies, roles and groups
        if not do_not_remove_iam_group:
            if verbose:
                logger.info("deleting IAM permissions %s" % sfn)
            iam = self.IAM(user_group_name)
            iam.delete_tibanna_iam(verbose=verbose, ignore_errors=ignore_errors)
        if purge_history:
            if verbose:
                logger.info("deleting all job files and history")
            item_list = dd_utils.get_items(DYNAMODB_TABLE, DYNAMODB_KEYNAME, 'Step Function', sfn, ['Log Bucket'])
            for item in item_list:
                jobid = item[DYNAMODB_KEYNAME]
                if 'Log Bucket' in item and item['Log Bucket']:
                    try:
                        keylist = retrieve_all_keys(jobid, item['Log Bucket'])
                    except Exception as e:
                        if 'NoSuchBucket' in str(e):
                            if verbose:
                                logger.info("log bucket %s missing... skip job %s" % (item['Log Bucket'], jobid))
                            continue
                    if verbose:
                        logger.info("deleting %d job files for job %s" % (len(keylist), jobid))
                    delete_keys(keylist, item['Log Bucket'])
                else:
                    if verbose:
                        logger.info("log bucket info missing.. skip job %s" % jobid)
            dd_utils.delete_items(DYNAMODB_TABLE, DYNAMODB_KEYNAME, item_list, verbose=verbose)
        if verbose:
            logger.info("Finished cleaning")

    def create_ami(self, build_from_scratch=True, source_image_to_copy_from=None, source_image_region=None,
                   ubuntu_base_image=None, make_public=False, replicate=False):
        args = dict()
        if build_from_scratch:
            # build from ubuntu 20.04 image and user data
            if ubuntu_base_image:
                args.update({'base_ami': ubuntu_base_image})
        else:
            # copy an existing image
            args.update({'userdata_file': ''})
            if source_image_to_copy_from:
                args.update({'base_ami': source_image_to_copy_from})
            if source_image_region:
                args.update({'base_region': source_image_region})

        return AMI(**args).create_ami_for_tibanna(make_public=make_public, replicate=replicate)



from tibanna.check_task import check_task
from tibanna.vars import AWS_REGION

config = {
    'function_name': 'check_task_awsem',
    'function_module': 'service',
    'function_handler': 'handler',
    'handler': 'service.handler',
    'region': AWS_REGION,
    'runtime': 'python3.8',
    'role': 'tibanna_lambda_init_role',
    'description': 'check status of AWSEM run by interegating appropriate files on S3 ',
    'timeout': 300,
    'memory_size': 256
}


def handler(event, context):
    return check_task(event)



from .vars import (
    AWS_REGION,
    AWS_ACCOUNT_NUMBER,
    SFN_TYPE,
    UPDATE_COST_LAMBDA_NAME
)
from .utils import create_tibanna_suffix


class StepFunctionCostUpdater(object):
    sfn_type = SFN_TYPE

    def __init__(self,
                 dev_suffix=None,
                 region_name=AWS_REGION,
                 aws_acc=AWS_ACCOUNT_NUMBER,
                 usergroup=None):
        self.dev_suffix = dev_suffix
        self.region_name = region_name
        self.aws_acc = aws_acc
        self.usergroup = usergroup

    @property
    def lambda_suffix(self):
        return create_tibanna_suffix(self.dev_suffix, self.usergroup)

    @property
    def lambda_arn_prefix(self):
        return "arn:aws:lambda:" + self.region_name + ":" + self.aws_acc + ":function:"

    @property
    def sfn_name(self):
        return 'tibanna_' + self.sfn_type + self.lambda_suffix + '_costupdater'

    @property
    def iam(self):
        from .iam_utils import IAM
        return IAM(self.usergroup)

    @property
    def sfn_role_arn(self):
        sfn_role_arn = "arn:aws:iam::" + self.aws_acc + ":role/" + \
                       self.iam.role_name('stepfunction')
        return sfn_role_arn

    @property
    def sfn_start_lambda(self):
        return 'Wait'

    @property
    def sfn_state_defs(self):
        state_defs = {
            "Wait": {
            "Type": "Wait",
            "Seconds": 43200, # Check every 12h
            "Next": "UpdateCostAwsem"
            },
            "UpdateCostAwsem": {
                "Type": "Task",
                "Resource": self.lambda_arn_prefix + UPDATE_COST_LAMBDA_NAME + self.lambda_suffix,
                "ResultPath": "$.done",
                "Next": "UpdateCostDone"
            },
            "UpdateCostDone": {
              "Type": "Choice",
              "Choices": [
                {
                  "Variable": "$.done.done",
                  "BooleanEquals": True,
                  "Next": "Done"
                }
              ],
              "Default": "Wait"
            },
            "Done": {
              "Type": "Pass",
              "End": True
            }
        }
        return state_defs

    @property
    def description(self):
        return "Update costs for a workflow run"

    @property
    def definition(self):
        return {
          "Comment": self.description,
          "StartAt": self.sfn_start_lambda,
          "States": self.sfn_state_defs
        }



from .vars import AWS_REGION, AWS_ACCOUNT_NUMBER
from .utils import create_tibanna_suffix


class StepFunctionUnicorn(object):
    lambda_error_retry_condition = {
        "ErrorEquals": [ "Lambda.ServiceException",
                         "Lambda.AWSLambdaException",
                         "Lambda.SdkClientException",
                         "Lambda.ResourceNotFoundException"],
        "IntervalSeconds": 60,
        "MaxAttempts": 6,
        "BackoffRate": 2
    }

    sfn_run_task_retry_conditions = [
        {
            "ErrorEquals": ["DependencyStillRunningException"],
            "IntervalSeconds": 600,
            "MaxAttempts": 10000,
            "BackoffRate": 1.0
        },
        {
            "ErrorEquals": ["EC2InstanceLimitWaitException"],
            "IntervalSeconds": 600,
            "MaxAttempts": 1008,  # 1 wk
            "BackoffRate": 1.0
        },
        lambda_error_retry_condition
    ]

    sfn_check_task_retry_conditions = [
        {
            "ErrorEquals": ["EC2StartingException"],
            "IntervalSeconds": 300,
            "MaxAttempts": 25,
            "BackoffRate": 1.0
        },
        {
            "ErrorEquals": ["StillRunningException"],
            "IntervalSeconds": 300,
            "MaxAttempts": 100000,
            "BackoffRate": 1.0
        },
        lambda_error_retry_condition
    ]

    def __init__(self,
                 dev_suffix=None,
                 region_name=AWS_REGION,
                 aws_acc=AWS_ACCOUNT_NUMBER,
                 usergroup=None):
        self.dev_suffix = dev_suffix
        self.region_name = region_name
        self.aws_acc = aws_acc
        self.usergroup = usergroup

    @property
    def lambda_suffix(self):
        return create_tibanna_suffix(self.dev_suffix, self.usergroup)

    @property
    def lambda_arn_prefix(self):
        return "arn:aws:lambda:" + self.region_name + ":" + self.aws_acc + ":function:"

    @property
    def sfn_name(self):
        return 'tibanna_unicorn' + self.lambda_suffix

    @property
    def iam(self):
        from .iam_utils import IAM
        return IAM(self.usergroup)

    @property
    def sfn_role_arn(self):
        sfn_role_arn = "arn:aws:iam::" + self.aws_acc + ":role/" + \
                       self.iam.role_name('stepfunction')
        return sfn_role_arn

    @property
    def sfn_start_lambda(self):
        return 'RunTaskAwsem'

    @property
    def sfn_state_defs(self):
        state_defs = {
            "RunTaskAwsem": {
                "Type": "Task",
                "Resource": self.lambda_arn_prefix + "run_task_awsem" + self.lambda_suffix,
                "Retry": self.sfn_run_task_retry_conditions,
                "Next": "CheckTaskAwsem"
            },
            "CheckTaskAwsem": {
                "Type": "Task",
                "Resource": self.lambda_arn_prefix + "check_task_awsem" + self.lambda_suffix,
                "Retry": self.sfn_check_task_retry_conditions,
                "End": True
            }
        }
        return state_defs

    @property
    def description(self):
        return "Start a workflow run on awsem and track it"

    @property
    def definition(self):
        return {
          "Comment": self.description,
          "StartAt": self.sfn_start_lambda,
          "States": self.sfn_state_defs
        }
